from scipy.ndimage import iterate_structure, generate_binary_structure
from scipy.ndimage import gaussian_filter, median_filter
from skimage.filters import threshold_otsu
from skimage.exposure import equalize_adapthist
import matplotlib.pyplot as plt
import numpy as np
from copy import deepcopy
import warnings


class ImageScalar:
    """
    Object represents a monochrome image.

    Attributes:

        im (np.ndarray[float]) - 2D array of pixel values

        shape (array like) - image dimensions

        mask (np.ndarray[bool]) - image mask

        labels (np.ndarray[int]) - segment ID mask

    """

    def __init__(self, im, labels=None):
        """
        Instantiate scalar image.

        Args:

            im (np.ndarray[float]) - 2D array of pixel values

           labels (np.ndarray[int]) - segment ID mask

        """
        self.im = im
        if im is not None:
            self.shape = im.shape[:2]
            self.mask = np.ones_like(self.im, dtype=bool)
        self.labels = labels

    def show(self,
             segments=True,
             cmap=None,
             vmin=0, vmax=1,
             figsize=(10, 10),
             ax=None,
             **kwargs):
        """
        Render image.

        Args:

            segments (bool) - if True, include cell segment contours

            cmap (matplotlib.colors.ColorMap or str) - colormap or RGB channel

            vmin, vmax (float) - bounds for color scale

            figsize (tuple) - figure size

            ax (matplotlib.axes.AxesSubplot) - if None, create axis

            kwargs: keyword arguments for add_contours

        Returns:

            fig (matplotlib.figures.Figure)

        """

        assert self.im is not None, 'Image not loaded.'

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        fig = plt.gcf()

        # show image in RGB format
        if type(cmap) == str:
            ind = 'rgb'.index(cmap)
            im = np.zeros((self.shape[0], self.shape[1], 3))
            im[:, :, ind] = self.im
            ax.imshow(im)

        # otherwise use specified colormap
        elif cmap is not None:
            ax.imshow(self.im, cmap=cmap, vmin=vmin, vmax=vmax)

        # otherwise show raw image
        else:
            ax.imshow(self.im)

        # add segment labels
        if segments and self.labels is not None:
            self.add_contours(ax, **kwargs)

        # remove axis
        ax.axis('off')

        return fig

    def add_contour(self, ax, mask, lw=1, color='r'):
        """ Adds border of specified contour. """
        ax.contour(mask, [0.5], linewidths=[lw], colors=[color])

    def add_contours(self, ax, lw=1, color='r', rasterized=False):
        """ Adds borders of all contours. """
        mask = self.labels > 0
        ctr = ax.contour(mask, [0.5], linewidths=[lw], colors=[color])
        if rasterized:
            for c in ctr.collections:
                c.set_rasterized(True)

    def gaussian_filter(self, sigma=(1., 1.)):
        """ Apply 2D gaussian filter. """
        self.im = gaussian_filter(self.im, sigma=sigma)

    def median_filter(self, radius=0, structure_dim=1):
        """ Apply 2D median filter. """
        struct = iterate_structure(generate_binary_structure(2, structure_dim), radius).astype(int)
        self.im = median_filter(self.im, footprint=struct)

    def set_mean_mask(self):
        """ Mask values below mean. """
        self.mask = np.zeros_like(self.im, dtype=bool)
        self.mask[self.im>=np.mean(self.im)] = True

    def set_otsu_mask(self):
        """ Mask values below otsu threahold. """
        threshold = threshold_otsu(self.im)
        self.mask = np.zeros_like(self.im, dtype=bool)
        self.mask[self.im>=threshold] = True

    def clahe(self, factor=8, clip_limit=0.01, nbins=256):
        """
        Run CLAHE on reflection-padded image.

        Args:


            factor (float or int) - number of segments per dimension

            clip_limit (float) - clip limit for CLAHE

            nbins (int) - number of grey-scale bins for histogram

        """

        # set kernel size as fraction of image size
        kernel_size = [int(np.ceil(s/factor)) for s in self.im.shape]

        # pad image with reflection about boundaries (circumvents artefacts)
        im_padded = np.pad(self.im, [(x,)*2 for x in kernel_size], mode='reflect')

        # apply CLAHE
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            im_equalized = equalize_adapthist(im_padded, kernel_size=kernel_size, clip_limit=clip_limit, nbins=nbins)

        # crop image
        self.im = im_equalized[tuple(slice(s, -s) for s in kernel_size)]

    def preprocess(self,
                   median_radius=2,
                   gaussian_sigma=(2, 2),
                   clip_limit=0.03,
                   clip_factor=20):
        """
        Preprocess image.

        Args:

            median_radius (int) - median filter size, px

            gaussian_sigma (tuple) - gaussian filter size, px std dev

            clip_limit (float) - CLAHE clip limit

            clip_factor (int) - CLAHE clip factor

        """
        self.median_filter(radius=median_radius)
        self.gaussian_filter(sigma=gaussian_sigma)
        self.clahe(clip_limit=clip_limit, factor=clip_factor)


class ImageMultichromatic(ImageScalar):
    """
    Object represents a multichromatic image.

    Attributes:

        im (np.ndarray[float]) - 2D array of pixel values in WHC format

    Inherited attributes:

        shape (array like) - image dimensions

        mask (np.ndarray[bool]) - image mask

        labels (np.ndarray[int]) - segment ID mask

    """

    def __init__(self, im, labels=None):
        """
        Instantiate RGB image.

        Args:

            im (np.ndarray[float]) - 2D array of pixel values

           labels (np.ndarray[int]) - segment ID mask

        """
        super().__init__(im, labels=labels)

    def get_channel(self, channel, copy=True):
        """
        Returns monochrome image of specified color channel.

        Args:

            channel (int) - desired channel

            copy (bool) - if True, instantiate from image copy

        Returns:

            image (ImageScalar) - monochrome image

        """

        if type(channel) == str:
            if 'ch' in channel:
                channel = int(channel.strip('ch'))

        if copy:
            monochrome = deepcopy(self.im[:, :, channel])
        else:
            monochrome = self.im[:, :, channel]

        return ImageScalar(monochrome, labels=self.labels)

    def to_RGB(self, channels_dict=None, copy=True):
        """
        Returns RGB image of specified color channels.

        Args:

            channels_dict (dict) - RGB channels keyed by channel index

            copy (bool) - if True, instantiate from image copy

        Returns:

            image (ImageMultichromatic) - RGB image

        """

        # default to first three channels as R, G, and B
        if channels_dict is None:
            channels_dict = dict(enumerate('rgb'))

        # copy image
        if copy:
            im = deepcopy(self.im)
        else:
            im = self.im

        # concatenate channels
        lookup = {v.lower().strip(): k for k, v in channels_dict.items()}
        channels = []
        for channel in 'rgb':
            if channel in lookup.keys():
                idx = lookup[channel]
                channels.append(im[:, :, idx])
            else:
                channels.append(np.zeros_like(im[:,:,0]))
        rgb = np.stack(channels, axis=-1)

        return ImageMultichromatic(rgb, labels=self.labels)



import numpy as np
import matplotlib.pyplot as plt

from ...visualization import default_figure


class ModelSelectionVisualization:
    """ Methods for visualizing model selection procedure. """

    def plot_models(self, panelsize=(3, 2), **kwargs):
        """ Plot model for each number of components. """
        ncols = self.max_num_components - 2
        fig, axes = plt.subplots(ncols=ncols, figsize=(panelsize[0]*ncols, panelsize[1]))
        for i, model in enumerate(self.models):
            model.plot_pdfs(ax=axes[i], **kwargs)
            if i == np.argmin(self.BIC):
                axes[i].set_title('SELECTED')

    @default_figure
    def plot_information_criteria(self,
                                  bic=True,
                                  aic=True,
                                  ax=None,
                                  **kwargs):
        """
        Plot information criteria versus number of components.

        Args:

            bic (bool) - include BIC scores

            aic (bool) - include AIC scores

            ax (matplotlib.axes.AxesSubplot) - if None, create axis

        Returns:

            fig (matplotlib.figure)

        """

        # plot AIC scores
        if aic:
            ind = np.argmin(self.AIC)
            ax.plot(self.num_components, self.AIC, '.-r', label='AIC', markersize=10)
            ax.scatter(self.num_components[ind], self.AIC[ind], s=150, facecolor='y', lw=1, marker=(5, 1), zorder=99)

        # plot BIC scores
        if bic:
            ind = np.argmin(self.BIC)
            ax.plot(self.num_components, self.BIC, '.-b', label='BIC', markersize=10)
            ax.scatter(self.num_components[ind], self.BIC[ind], s=150, facecolor='y', lw=1, marker=(5, 1), zorder=99)

        # format axes
        ax.set_xlabel('Number of components in mixture')
        ax.set_ylabel('Information Criteria')
        ax.set_xticks(self.num_components)
        ax.set_yticks([])
        ax.legend(frameon=False, bbox_to_anchor=(0., 1.02, 1., .102), loc='upper center', ncol=2, mode="center", borderaxespad=0.)
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)



import numpy as np
from collections import Counter


def resample_uniformly(x, y, size=None, cutoff=None):
    """
    Resample X and Y uniformly in X.

    Args:

        x, y (np.ndarray[float]) - original samples

        size (int) - number of uniform samples

        cutoff (int) - upper bound for samples (quantile, 0 to 100)

    Returns:

        x, y (np.ndarray[float]) - resampled s.t. x is uniformly distributed

    """

    if size is None:
        size = x.size

    # sort values
    sort_ind = np.argsort(x)
    xx, yy = x[sort_ind], y[sort_ind]

    # apply threshold on upper bound
    if cutoff is not None:
        threshold = np.percentile(xx, cutoff)
    else:
        threshold = xx.max()+1

    # get unique x values
    xunique = np.unique(xx)

    # filter points below threshold
    below_threshold = (xx<=threshold)
    xx, yy = xx[below_threshold], yy[below_threshold]

    # get probabilities
    x_to_count = np.vectorize(Counter(xx).get)

    # get intervals
    intervals = np.diff(xunique)
    unique_below_threshold = (xunique[:-1]<=threshold)
    intervals = intervals[unique_below_threshold]

    # assign probabilities
    x_to_cumul = np.vectorize(dict(zip(xunique[:-1][unique_below_threshold], intervals/intervals.sum())).get)
    p = x_to_cumul(xx)/x_to_count(xx)
    p[np.isnan(p)] = 0

    # generate sample
    sample_ind = np.random.choice(np.arange(xx.size), size=size, p=p)
    xu, yu = xx[sample_ind], yy[sample_ind]

    return xu, yu



import numpy as np
from scipy.stats import binned_statistic
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt

from ...visualization import *

from .timeseries import plot_mean, plot_mean_interval
from .timeseries import smooth


class CorrelationProperties:

    # fraction of cells used to determine window size
    window_frac = 0.01

    @property
    def num_fluctuations(self):
        """ Number of pairwise fluctuations. """
        return len(self.d_ij)

    @property
    def ubound(self):
        """ Upper bound on pairwise distances. """
        return np.percentile(self.d_ij, 99)

    @property
    def num_measurements(self):
        """ Estimated number of measurements. """
        return np.sqrt(self.d_ij.size)

    @property
    def window(self):
        """ Window size. """
        return int(np.round((self.num_measurements * self.window_frac) ** 2))

    @property
    def d_av(self):
        """ Smoothed moving average distances. """
        return smooth(self.d_ij, self.window)

    @property
    def C_av(self):
        """ Smoothed moving average correlation. """
        return smooth(self.C_ij, self.window)

    @property
    def characteristic_length(self):
        """ Characteristic length over which correlations decay. """
        try:
            length = CharacteristicLength(self).characteristic_length
        except:
            length = None
        return length


class CorrelationVisualization:
    """ Visualization methods for SpatialCorrelation. """

    @default_figure
    def histogram_distances(self, ax=None, **kwargs):
        """ Plot histogram of pairwise distances between measurements. """
        _ = ax.hist(self.d_ij, **kwargs)
        ax.set_ylabel('Num. pairs')
        ax.set_xlabel('Separation distance')

    @default_figure
    def histogram_fluctuations(self, ax=None, **kwargs):
        """ Plot histogram of pairwise fluctuations between measurements. """
        _ = ax.hist(self.C_ij, **kwargs)
        ax.set_ylabel('Num. pairs')
        ax.set_xlabel('Fluctuation')

    @default_figure
    def visualize(self,
                  null_model=False,
                  scatter=False,
                  confidence=False,
                  zero=True,
                  smooth=True,
                  ma_type='sliding',
                  window_size=500,
                  resolution=100,
                  nbootstraps=50,
                  color='k',
                  max_distance=None,
                  ax=None):
        """
        Plot pairwise normalized fluctuations versus pairwise distances.

        Args:

            null_model (bool) - if True, shuffle d_ij vector

            scatter (bool) - if True, show individual markers

            confidence (bool) - if True, include confidence interval

            zero (bool) - if True, include zero correlation line for reference

            smooth (bool) - if True, apply smoothing to moving average

            ma_type (str) - type of average used, either sliding, binned, or savgol

            window_size (int) - size of window

            resolution (int) - sampling interval

            nbootstraps (int) - number of bootstrap samples for confidence interval

            color (str) - color used for confidence interval

            max_distance (float) - largest pairwise distance included

            ax (mpl.axes.AxesSubplot) - if None, create figure

        Returns:

            ax (mpl.axes.AxesSubplot)

        """

        d = self.d_ij
        C = self.C_ij

        # if null_model is True, randomly shuffle d_ij vector
        if null_model:
            d = np.random.choice(d, self.num_fluctuations, replace=False)

            # sort by distance
            ind = np.argsort(d)
            d = d[ind]
            C = C[ind]

        # filter by max_distance
        if max_distance is None:
            max_distance = self.ubound
        ind = (d<=max_distance)
        C = C[ind]
        d = d[ind]

        # get smoothing arguments
        ma_kw=dict(ma_type=ma_type, window_size=window_size, resolution=resolution)

        # plot moving average
        plot_mean(ax=ax, x=d, y=C,
                  line_color=color,
                  line_width=1,
                  line_alpha=1,
                  markersize=2,
                  smooth=smooth,
                  **ma_kw)

        # plot confidence interval for moving average
        if confidence:
            plot_mean_interval(ax=ax, x=d, y=C,
                               confidence=95,
                               color=color,
                               alpha=0.35,
                               nbootstraps=nbootstraps,
                               **ma_kw)

        # plot markers
        if scatter:
            ax.scatter(d, C, alpha=0.2, color='grey', linewidth=0)

        # plot zero reference line
        if zero:
            ax.plot([0, max_distance], [0, 0], '-r', linewidth=1, alpha=0.25)

        # format
        ymin, ymax = -0.15, 0.7
        ax.set_ylim(ymin, ymax), ax.set_yticks([-0., 0.2, 0.4, .6])
        ax.set_xlim(0, max_distance)
        ax.set_xlabel('Pairwise distance')
        ax.set_ylabel('Correlation')


class SpatialCorrelation(CorrelationProperties, CorrelationVisualization):
    """
    Container for correlations between 1-D timeseries.

    Attributes:

        d_ij (np array) - pairwise separation distances between measurements

        C_ij (np array) - normalized pairwise fluctuations between measurements

    """

    def __init__(self, d_ij=None, C_ij=None):
        """
        Instantiate container for correlations between 1-D timeseries.

        Args:

            d_ij (np.ndarray) - pairwise separation distances

            C_ij (np.ndarray) - normalized pairwise fluctuations

        """

        if d_ij is None:
            d_ij = np.empty(0)
        if C_ij is None:
            C_ij = np.empty(0)

        ind = np.argsort(d_ij)
        self.d_ij = d_ij[ind]
        self.C_ij = C_ij[ind]

    def __add__(self, correlation):
        """
        Concatenate current instance with a second SpatialCorrelation instance.

        Args:

            correlation (SpatialCorrelation)

        Returns:

            self (SpatialCorrelation) - updated correlations

        """

        # concatenate distanced and fluctuations
        d_ij = np.hstack((self.d_ij, correlation.d_ij))
        C_ij = np.hstack((self.C_ij, correlation.C_ij))

        # sort by distance
        ind = np.argsort(d_ij)
        self.d_ij = d[ind]
        self.C_ij = C_ij[ind]

        return self


class CharacteristicLength:
    """
    Class for determining the characteristic length over which correlations decay.
    """

    def __init__(self, correlation, fraction_of_max=0.01):
        """
        Args:

            correlation (SpatialCorrelation)

            fraction_of_max (float) - fraction of peak correlation used to fit exponential decay

        """

        self.fraction_of_max = fraction_of_max

        # extract correlation decay dynamics
        xmax, x, y = self.extract_decay(correlation, fraction_of_max)
        self.xmax = xmax
        self.x = x
        self.y = y

        # fit exponential decay
        popt, perr = self.fit(self.x_normed, self.y, self.model)
        self.popt = popt
        self.perr = perr

    @property
    def x_normed(self):
        """ Distance vector normalized by maximum value. """
        return self.x / self.xmax

    @property
    def yp(self):
        """ Predicted correlation values. """
        return self.model(self.x_normed, *self.popt)

    @property
    def characteristic_length(self):
        """ Characteristic decay length. """
        return self.xmax / self.popt[1]

    @staticmethod
    def extract_decay(correlation, fraction_of_max):
        """ Extract decay. """

        x = correlation.d_av
        y = correlation.C_av
        ymin = fraction_of_max * y.max()
        max_idx = (y < ymin).nonzero()[0][0]

        return x[max_idx], x[0:max_idx], y[0:max_idx]

    @staticmethod
    def model(x, a, b):
        """ Exponential decay model. """
        return a * np.exp(-b*x)

    @staticmethod
    def fit(x, y, model):
        """ Fit exponential decay model to decay vectors. """
        popt, pcov = curve_fit(model, x, y, p0=(1., 1.))
        perr = np.sqrt(np.diag(pcov))
        return popt, perr

    def plot_measured(self, ax, **kwargs):
        """ Plot measured correlation decay. """
        ax.plot(self.x, self.y, label='Measured', **kwargs)

    def plot_fit(self, ax, **kwargs):
        """ Plot model fit. """
        ax.plot(self.x, self.yp, label='Fit', **kwargs)

    @default_figure
    def plot(self, ax=None, **kwargs):
        """ Plot measured correlation decay alongside model fit. """
        self.plot_measured(ax, color='k', **kwargs)
        self.plot_fit(ax, color='r', **kwargs)
        ax.set_ylabel('Correlation')
        ax.set_xlabel('Distance')




from ..classification import BivariateMixtureClassifier
from .univariate import UnivariateModelSelection


class BivariateModelSelection(UnivariateModelSelection):
    """ Bivariate extension for model selection. """

    @staticmethod
    def load_model(path):
        """ Load model from <path> directory. """
        return BivariateMixtureClassifier.load(path)

    @staticmethod
    def fit_model(values, num_components, num_labels, **kwargs):
        """ Fit model with specified number of components. """
        return BivariateMixtureClassifier(values,
                                        num_components=num_components,
                                        num_labels=num_labels,
                                        **kwargs)



import numpy as np
import matplotlib.pyplot as plt
from collections import Counter
from scipy.stats import norm
from sklearn.mixture import GaussianMixture

from .visualization import MixtureVisualization


class MixtureIO:

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["values"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.values = None


class MixtureProperties:
    """ Properties for guassian mixture models. """

    @property
    def means(self):
        """ Mean value of each component. """
        return self.means_[:, 0].ravel()

    @property
    def stds(self):
        """ Standard deviation of each component. """
        return np.sqrt(self.covariances_[:, 0].ravel())

    @property
    def lbound(self):
        """ Lower bound of support. """
        return np.percentile(self.values[:, 0], q=0.1)

    @property
    def ubound(self):
        """ Upper bound of support. """
        return np.percentile(self.values[:, 0], q=99.9)

    @property
    def bounds(self):
        """ Low and upper bounds of support. """
        return np.percentile(self.values[:, 0], q=[.1, 99.9])

    @property
    def support(self):
        """ Distribution support. """
        return np.linspace(self.lbound, self.ubound, num=100)

    @property
    def num_components(self):
        """ Number of model components. """
        return self.n_components

    @property
    def log_likelihood(self):
        """ Maximized log likelihood. """
        return self.score(self.values) * self.num_samples

    @property
    def BIC(self):
        """ BIC score. """
        return self.bic(self.values)

    @property
    def AIC(self):
        """ AIC score. """
        return self.aic(self.values)

    @property
    def scale_factor(self):
        """ Scaling factor for log-transformed support. """
        return np.exp(self.support)

    @property
    def support_size(self):
        """ Size of support. """
        return self.support.size

    @property
    def num_samples(self):
        """ Number of samples. """
        return self.values.shape[0]

    @property
    def components(self):
        """ Individual model components. """
        get_params = lambda i: (self.means_[i], np.sqrt(self.covariances_[i]))
        build_component = lambda i: norm(*get_params(i))
        return [build_component(i) for i in range(self.n_components)]

    @property
    def pdf(self):
        """ Gaussian Mixture PDF. """
        pdf = np.zeros(self.support_size)
        for i in range(self.n_components):
            pdf += self.get_component_pdf(i)
        return pdf

    @property
    def component_pdfs(self):
        """ Returns stacked array of component PDFs. """
        return self._component_pdfs()

    def _component_pdfs(self, weighted=True):
        """ Returns stacked array of component PDFs over support. """
        build_pdf = lambda idx: self.get_component_pdf(idx, weighted)
        return np.stack([build_pdf(idx) for idx in range(self.n_components)])


class UnivariateMixture(GaussianMixture,
                        MixtureIO,
                        MixtureProperties,
                        MixtureVisualization):
    """
    Univariate Gaussian mixture model.

    Attributes:

        values (array like) - values to which model was fit

    Inherited attributes:

        See sklearn.mixture.GaussianMixture

    """

    dim = 1

    def __init__(self, *args, values=None, **kwargs):
        """ Instantiate Gaussian mixture model. """
        self.values = values
        GaussianMixture.__init__(self, *args, **kwargs)

    def __repr__(self):
        """ Displays summary of mixture components. """
        return self.summary

    @classmethod
    def from_logsample(cls, sample,
                    n=3,
                    max_iter=10000,
                    tol=1e-8,
                    covariance_type='diag',
                    n_init=10):
        """ Instantiate from log-transformed sample. """

        if len(sample.shape) != cls.dim + 1:
            sample = sample.reshape(-1, cls.dim)

        model = cls(n, values=sample,
                    covariance_type=covariance_type,
                    tol=tol,
                    n_init=n_init,
                    max_iter=max_iter)

        model = model.fit(sample)

        return model

    @classmethod
    def from_sample(cls, sample, n, **kwargs):
        """ Instantiate from log-normally distributed sample. """
        return cls.from_logsample(np.log(sample), n=n, **kwargs)

    @classmethod
    def from_parameters(cls, mu, sigma, weights=None, values=None, **kwargs):
        """ Instantiate model from parameter vectors. """

        model = cls(len(mu), values=values, covariance_type='diag', **kwargs)
        model.fit(np.random.rand(50, 1))  # train on false data

        if type(mu) == list:
            mu = np.array(mu).reshape(-1, 1)

        if type(sigma) == list:
            sigma = np.array(sigma).reshape(-1, 1)

        elif type(sigma) in (int, float, np.float64, np.int64):
            sigma = np.ones((len(mu), 1), dtype=float) * sigma

        if weights is None:
            weights = np.ones(len(mu), dtype=float) / len(mu)
        elif type(weights) == list:
            weights = np.array(weights)

        model.weights_ = weights
        model.means_ = mu
        model.covariances_ = sigma
        precisions = np.linalg.inv(np.diag(sigma.ravel()))
        model.precisions_ = np.diag(precisions).reshape(-1, 1)
        cholesky_precision = np.linalg.cholesky(precisions)
        model.precisions_cholesky_ = np.diag(cholesky_precision).reshape(-1, 1)
        model.lower_bound_ = model.score(model.values)

        if values is None:
            model.values = model.logsample(1000)

        return model

    def logsample(self, N):
        """ Returns <N> samples of log-transformed variable. """
        logsamples, labels = GaussianMixture.sample(self, N)
        return logsamples

    def sample(self, N):
        """ Returns <N> samples of variable. """
        return np.exp(self.logsample(N))

    def sample_component(self, component_idx, N):
        """ Returns <N> log-transformed samples from indexed component. """
        return self.components[component_idx].rvs(N)

    def multi_logsample(self, N, m=10):
        """
        Returns <N> log-transformed samples as well as <N> log-transformed samples averaged over <m> other samples from the same component.
        """

        component_idxs = np.random.choice(np.arange(self.n_components), size=N, p=self.weights_)
        N_per_component = Counter(component_idxs)

        data = {}
        for component_idx, num_samples in N_per_component.items():
            subsample = self.sample_component(component_idx, num_samples)
            neighbor_idxs = np.random.randint(0, num_samples, size=(num_samples, m))
            neighbor_average = subsample[neighbor_idxs].mean(axis=1)
            data[component_idx] = (subsample, neighbor_average)
        data = np.hstack(data.values())

        return data.T

    def multi_sample(self, N, m=10):
        """
        Returns <N> samples as well as <N> samples averaged over <m> other samples from the same component.
        """
        return np.exp(self.multi_logsample(N, m))

    def get_component_pdf(self, idx, weighted=True):
        """ Returns PDF for indexed component. """

        pdf = self.components[idx].pdf(self.support).reshape(self.support_size)

        if weighted:
            pdf *= self.weights_[idx]

        return pdf

    def estimate_required_samples(self, SNR=5.):
        """
        Returns minimum number of averaged samples required to achieve the specified signal to noise (SNR) ratio.
        """

        yy, xx = np.meshgrid(*(np.arange(self.n_components),)*2)
        delta_mu = np.abs(self.means[xx]-self.means[yy])
        max_sigma = np.stack((self.stds[xx], self.stds[yy])).max(axis=0)

        idx = np.triu_indices(self.n_components, k=1)
        min_num_samples = (SNR*(max_sigma[idx]/delta_mu[idx])**2).max()

        return min_num_samples



import numpy as np
from scipy.spatial import Delaunay
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection


class AlphaVisualization:
    """ Visualization methods for AlphaShapes class. """


    def plot_boundary(self, color='k', ax=None, **kwargs):
        """
        Plot boundary of concave hull as a collection of line objects.

        Args:

            color (str or RGBA) - path color

            ax (matplotlib.axes.AxesSubplot) - if None, create figure

            kwargs: keyword arguments for matplotlib.LineCollection

        """

        # create figure
        if ax is None:
            fig, ax = plt.subplots(figsize=(2, 2))
            xmin, ymin = self.points.min(axis=0)
            xmax, ymax = self.points.max(axis=0)
            ax.set_xlim(xmin, xmax)
            ax.set_ylim(ymin, ymax)
            ax.axis('off')

        # create line collection and add it to the axis
        lines = LineCollection(self.boundary, colors=color, **kwargs)
        ax.add_collection(lines)

        # format axis
        ax.set_aspect(1)

        return ax


class AlphaShapes(AlphaVisualization):
    """
    Object for computing the concave hull (alpha shape) of a set of points.

    Attributes:

        points (np.ndarray[float]) - points, shape (n,2)

        alpha (float) - alpha value

        only_outer (bool) - if True, only keep outer border

        edge_indices (np.ndarray[int]) - (i,j) pairs representing edges of the alpha-shape. Indices (i,j) index the points array.

    References:

    stackoverflow: questions/50549128/boundary-enclosing-a-given-set-of-points

    """

    def __init__(self, points, alpha, only_outer=True):
        """
        Instantiate alpha shape object.

        Args:

            points (np.ndarray[float]) - points, shape (n,2)

            alpha (float) - alpha value

            only_outer (bool) - if True, only keep outer border

        """

        self.points = points
        self.alpha = alpha
        self.only_outer = only_outer
        edge_indices = self._alpha_shape(points, alpha, only_outer=only_outer)
        self.edge_indices = np.array(list(edge_indices))

    def __call__(self, xy):
        """ Test whether each point in <xy> lies within the alpha shape. """
        f = np.vectorize(lambda x, y: self._is_inside(x, y, self.points, self.edges))
        return f(xy.T)

    @property
    def boundary(self):
        """ Boundary line segments. """
        return self.points[self.edge_indices]

    @staticmethod
    def _alpha_shape(points, alpha, only_outer=True):
        """
        Compute the concave hull (alpha shape) of a set of points.

        Args:

            points (np.ndarray[float]) - points, shape (n,2)

            alpha (float) - alpha value

            only_outer (bool) - if True, only keep outer border

        Returns:

            boundary (list of tuples) - Set of (i,j) pairs representing edges of the alpha-shape. Indices (i,j) index the points array.

        References:

            https://stackoverflow.com/questions/50549128/boundary-enclosing-a-given-set-of-points

        """
        assert points.shape[0] > 3, "Need at least four points"

        def add_edge(edges, i, j):
            """ Add a line between the i-th and j-th points if it's not already in the list. """
            if (i, j) in edges or (j, i) in edges:
                # already added
                assert (j, i) in edges, "Can't go twice over same directed edge right?"
                if only_outer:
                    # if both neighboring triangles are in shape, it's not a boundary edge
                    edges.remove((j, i))
                return
            edges.add((i, j))

        tri = Delaunay(points)
        edges = set()
        # Loop over triangles:
        # ia, ib, ic = indices of corner points of the triangle
        for ia, ib, ic in tri.vertices:
            pa = points[ia]
            pb = points[ib]
            pc = points[ic]
            # Computing radius of triangle circumcircle
            # www.mathalino.com/reviewer/derivation-of-formulas/derivation-of-formula-for-radius-of-circumcircle
            a = np.sqrt((pa[0] - pb[0]) ** 2 + (pa[1] - pb[1]) ** 2)
            b = np.sqrt((pb[0] - pc[0]) ** 2 + (pb[1] - pc[1]) ** 2)
            c = np.sqrt((pc[0] - pa[0]) ** 2 + (pc[1] - pa[1]) ** 2)
            s = (a + b + c) / 2.0
            area = np.sqrt(s * (s - a) * (s - b) * (s - c))
            circum_r = a * b * c / (4.0 * area)
            if circum_r < alpha:
                add_edge(edges, ia, ib)
                add_edge(edges, ib, ic)
                add_edge(edges, ic, ia)
        return edges

    @staticmethod
    def _is_inside(x, y, points, edges, eps=1.0e-10):
        """
        Check if point (<x>, <y>) lies within the alpha shape defined by <points> and <edges>.
        """
        intersection_counter = 0
        for i, j in edges:
            assert abs((points[i,1]-y)*(points[j,1]-y)) > eps, 'Need to handle these end cases separately'
            y_in_edge_domain = ((points[i,1]-y)*(points[j,1]-y) < 0)
            if y_in_edge_domain:
                upper_ind, lower_ind = (i,j) if (points[i,1]-y) > 0 else (j,i)
                upper_x = points[upper_ind, 0]
                upper_y = points[upper_ind, 1]
                lower_x = points[lower_ind, 0]
                lower_y = points[lower_ind, 1]

                # is_left_turn predicate is evaluated with: sign(cross_product(upper-lower, p-lower))
                cross_prod = (upper_x - lower_x)*(y-lower_y) - (upper_y - lower_y)*(x-lower_x)
                assert abs(cross_prod) > eps, 'Need to handle these end cases separately'
                point_is_left_of_segment = (cross_prod > 0.0)
                if point_is_left_of_segment:
                    intersection_counter = intersection_counter + 1
        return (intersection_counter % 2) != 0



from os import getcwd
from argparse import ArgumentParser, ArgumentTypeError


# ======================== ARGUMENT TYPE CASTING ==============================

def str2bool(arg):
     """ Convert <arg> to boolean. """
     if arg.lower() in ('yes', 'true', 't', 'y', '1'):
          return True
     elif arg.lower() in ('no', 'false', 'f', 'n', '0'):
          return False
     else:
          raise ArgumentTypeError('Boolean value expected.')


# ======================== ARGUMENT PARSING ===================================


class RunArguments(ArgumentParser):
     """ Argument handler for run scripts. """

     def __init__(self, **kwargs):
          super().__init__(**kwargs)
          self.add_arguments()
          self.parse()

     def __getitem__(self, key):
          """ Returns <key> argument value. """
          return self.args[key]

     def add_arguments(self):
          """ Add arguments. """

          # add position argument for path
          self.add_argument(
               'path',
               nargs='?',
               default=getcwd())

          # add keyword argument for saving measurement data
          self.add_argument(
               '-S', '--save_data',
               help='Save measurement (training) data.',
               type=str2bool,
               default=False,
               required=False)

          # add keyword argument for globally training annotator
          self.add_argument(
               '-G', '--train_globally',
               help='Train global annotator.',
               type=str2bool,
               default=True,
               required=False)

     def parse(self):
          """ Parse arguments. """
          self.args = vars(self.parse_args())


class SweepArguments(RunArguments):
     """ Argument handler for parameter sweeps. """

     def add_arguments(self):
          """ Add arguments. """

          super().add_arguments()

          # add keyword argument for minimum fluorescence ambiguity
          self.add_argument('--min_ambiguity',
                              help='Minimum fluorescence ambiguity.',
                              type=float,
                              default=0.1,
                              required=False)

          # add keyword argument for maximum fluorescence ambiguity
          self.add_argument('--max_ambiguity',
                              help='Maximum fluorescence ambiguity.',
                              type=float,
                              default=1.,
                              required=False)

          # add keyword argument for number of fluorescence ambiguities
          self.add_argument('-a', '--num_ambiguities',
                              help='Number of fluorescence ambiguities.',
                              type=int,
                              default=16,
                              required=False)

          # add keyword argument for number of fluorescence replicates
          self.add_argument('-n', '--num_replicates',
                              help='Fluorescence replicates per simulation.',
                              type=int,
                              default=1,
                              required=False)

          # add keyword argument for sweep density
          self.add_argument('-lr', '--logratio',
                              help='Weight edges by logratio.',
                              type=str2bool,
                              default=True,
                              required=False)

          # add keyword argument for estimated run time
          self.add_argument('-w', '--walltime',
                              help='Estimated run time.',
                              type=int,
                              default=24,
                              required=False)

          # add keyword argument for number of cores
          self.add_argument('-c', '--cores',
                              help='Number of cores.',
                              type=int,
                              default=4,
                              required=False)

          # add keyword argument for memory usage
          self.add_argument('-m', '--memory',
                              help='Memory usage.',
                              type=int,
                              default=4,
                              required=False)

          # add keyword argument for project allocation
          self.add_argument('-A', '--allocation',
                              help='Project allocation.',
                              type=str,
                              default='p30653',
                              required=False)



__author__ = 'Sebastian Bernasek'

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap, ListedColormap


def build_transparent_cmap(fg, bg='w'):
    """
    Construct colormap with linear scaled transparency between <bg> and <fg> colors .
    """
    base_map = LinearSegmentedColormap.from_list(fg, [bg, fg])
    cmap = base_map(np.arange(base_map.N))
    cmap[:,-1] = np.linspace(0, 1, base_map.N)
    cmap = ListedColormap(cmap)
    return cmap


# labels
labelpad = 1
labelsize = 7

# tick labels
tickpad = 1
ticklabelsize = 6

# font params
plt.rcParams['font.size'] = ticklabelsize
plt.rcParams['font.family'] = 'sans-serif'
plt.rcParams['font.sans-serif'] = ['Arial']
plt.rcParams['pdf.fonttype'] = 42
plt.rcParams['ps.fonttype'] = 42

# axes params
plt.rcParams['axes.labelpad'] = labelpad
plt.rcParams['axes.labelsize'] = labelsize
plt.rcParams['axes.labelcolor'] = '000000'
plt.rcParams['axes.linewidth'] = 1
plt.rcParams['axes.edgecolor'] = '000000'

#xtick params
plt.rcParams['xtick.major.width'] = 1
plt.rcParams['xtick.minor.width'] = 1
plt.rcParams['xtick.major.size'] = 3
plt.rcParams['xtick.minor.size'] = 1
plt.rcParams['xtick.major.pad'] = tickpad
plt.rcParams['xtick.labelsize'] = ticklabelsize
plt.rcParams['xtick.direction'] = 'out'
plt.rcParams['xtick.color'] = '000000'

# ytick params
plt.rcParams['ytick.major.width'] = 1
plt.rcParams['ytick.minor.width'] = 1
plt.rcParams['ytick.major.size'] = 3
plt.rcParams['ytick.minor.size'] = 1
plt.rcParams['ytick.major.pad'] = tickpad
plt.rcParams['ytick.labelsize'] = ticklabelsize
plt.rcParams['ytick.direction'] = 'out'
plt.rcParams['ytick.color'] = '000000'

# save params
plt.rcParams['savefig.bbox'] = 'tight'
plt.rcParams['savefig.pad_inches'] = 0.1
plt.rcParams['savefig.transparent'] = True



from sklearn.cluster import k_means
from scipy import interpolate
import numpy as np
import matplotlib.pyplot as plt


class RBF:
    """
    Radial basis function fit to scalar field.
    """

    def __init__(self, xy, z, smooth=1, function='multiquadric'):
        """
        Instantiate RBF object.self

        Args:

            xy (2D np.ndarray[float]) - node coordinates (2xN)

            z (1D np.ndarray[float]) - scalar values

            smooth (float) - RBF smoothing parameter

            function (str) - RBF kernel

        """
        self.xy = xy
        self.z = z
        self.rbf = self.fit(smooth, function)
        self.evaluate()
        self.cluster()

    def __call__(self, *args):
        return self.rbf(*args)

    def fit(self,
            smooth=1,
            function='multiquadric'):
        """ Fit RBF """
        rbf = interpolate.rbf.Rbf(*self.xy, self.z, smooth=smooth, function=function)
        return rbf

    def evaluate(self):
        self.zi = self.__call__(*self.xy)

    def cluster(self, n=3):
        centroids, labels, inertia = k_means(self.zi.reshape(-1, 1), n)
        self.labels = np.empty_like(labels, dtype=int)
        self.thresholds = np.empty_like(centroids.flatten())
        indices = np.argsort(centroids.flatten())
        for label, ind in enumerate(indices):
            self.labels[labels==ind] = label
            self.thresholds[label] = self.zi[labels==ind].max()
        self.centroids = centroids[indices]

    def plot_points(self,
                    ax,
                    s=10,
                    lw=0,
                    cmap=None):
        if cmap is None:
            cmap = plt.cm.Reds
        y, x = self.xy
        vmin, vmax = self.z.min(), self.z.max()
        ax.scatter(x, y, c=self.zi, vmin=vmin, vmax=vmax, cmap=cmap, s=s, linewidths=lw)
        ax.axis('off')

    def show(self):
        fig, ax = plt.subplots(figsize=(2, 2))
        ax.plot(self.z, self.zi, '.k')
        ax.set_xlabel('Data')
        ax.set_ylabel('Interpolated')



from .model_selection import UnivariateModelSelection, BivariateModelSelection
from .spatial import WeightedGraph
from .annotation import Annotation
from .spatial.concurrency import ConcurrencyLabeler
from .visualization import CloneBoundaries
from .labelers import CelltypeLabeler



import numpy as np
from matplotlib.tri import Triangulation
from scipy.spatial import ConvexHull


class LocalTriangulation(Triangulation):
    """
    Triangulation with edge distance filter.

    Attributes:

        edge_list (np.ndarray[int]) - (from, to) node pairs

        edge_lengths (np.ndarray[float]) - euclidean length of each edge

    """

    def __init__(self, *args, **kwargs):
        """
        Instantiate triangulation.

        Args:

            x, y (np.ndarray[float]) - spatial coordinates

        """

        # call matplotlib.tri.Triangulation instantiation
        super().__init__(*args, **kwargs)

        # compile edges
        edge_list = self.compile_edge_list()
        edge_lengths = self.evaluate_edge_lengths(edge_list, self.x, self.y)

        # store edges
        self.edge_list = edge_list
        self.edge_lengths = edge_lengths

    @property
    def size(self):
        """ Number of points. """
        return self.x.size

    @property
    def radii(self):
        """ Radius. """
        return np.sqrt(self.x**2 + self.y**2)

    @property
    def angles(self):
        """ Angle on [0, 2p] interval. """
        return np.arctan2(self.y, self.x) + np.pi

    @property
    def hull(self):
        """ Convex hull. """
        return ConvexHull(np.vstack((self.x, self.y)).T)

    @property
    def num_triangles(self):
        """ Number of triangles. """
        return self.triangles.shape[0]

    @property
    def nodes(self):
        """ All nodes. """
        return np.unique(self.triangles)

    @property
    def edges(self):
        """ Filtered edges. """
        return self.filter_outliers(self.nodes, self.edge_list, self.edge_lengths)
        #return self.filter_edges(self.nodes, self.edge_list, self.edge_lengths)
        #return self.filter_hull(self.edge_list)
        #return self.filter_longest_edge(self.edge_list, self.edge_lengths)

    def compile_edge_list(self):
        """ Returns list of (node_from, node_to) tuples. """
        edges = []
        for i in range(3):
            edges += list(zip(self.triangles[:, i], self.triangles[:,(i+1)%3]))
        return np.array(edges)

    @staticmethod
    def evaluate_edge_lengths(edge_list, x, y):
        """ Returns 1D array of edge lengths. """
        dx, dy = x[edge_list], y[edge_list]
        return np.sqrt(np.diff(dx, axis=1)**2 + np.diff(dy, axis=1)**2).reshape(-1)

    @staticmethod
    def find_disconnected_nodes(nodes, edges):
        """ Returns boolean array of nodes not included in edges. """
        return nodes[~np.isin(nodes, np.unique(edges))]

    @staticmethod
    def find_first_edge(edges, node):
        """ Returns index of first edge containing <node>. """
        return (edges==node).any(axis=1).nonzero()[0][0]

    @classmethod
    def filter_edges(cls, nodes, edges, lengths, max_length=0.1):
        """ Returns all edges less than <max_length>, with at least one edge containing each node. """

        # sort edges
        sort_indices = np.argsort(lengths)
        edges = edges[sort_indices]
        lengths = lengths[sort_indices]

        mask = (lengths <= max_length)
        rejected, accepted = edges[~mask], edges[mask]

        # find disconnected nodes
        disconnected = cls.find_disconnected_nodes(nodes, accepted)

        # add shortest edge for each disconnected node
        if disconnected.size > 0:
            f = np.vectorize(lambda node: cls.find_first_edge(rejected, node))
            connecting = rejected[f(disconnected)]
            accepted = np.vstack((accepted, connecting))

        return accepted

    @classmethod
    def filter_outliers(cls, nodes, edges, lengths):
        """
        Returns all edges whose lengths are not outliers, with at least one edge containing each node.
        """

        # sort edges
        sort_indices = np.argsort(lengths)
        edges = edges[sort_indices]
        lengths = lengths[sort_indices]

        mask = ~cls.is_outlier(lengths)

        rejected, accepted = edges[~mask], edges[mask]

        # find disconnected nodes
        disconnected = cls.find_disconnected_nodes(nodes, accepted)

        # add shortest edge for each disconnected node
        if disconnected.size > 0:
            f = np.vectorize(lambda node: cls.find_first_edge(rejected, node))
            connecting = rejected[f(disconnected)]
            accepted = np.vstack((accepted, connecting))

        return accepted

    def filter_hull(self, edges):
        """ Returns all edges not on the convex hull. """
        hull_edges = np.sort(self.hull.simplices, axis=1)
        on_hull = np.isin(np.sort(edges, axis=1), hull_edges).all(axis=1)
        return edges[~on_hull]

    def filter_longest_edge(self, edges, edge_lengths):
        """ Returns all edges except the longest edge in each triangle. """
        accepted_edges = []
        for tri in range(self.num_triangles):
            ind = np.argsort(edge_lengths[tri::self.num_triangles])[:-1]
            accepted_edges.append(edges[tri::self.num_triangles][ind])
        return np.vstack(accepted_edges)

    @staticmethod
    def is_outlier(points, threshold=3.):
        """
        Returns a boolean array with True if points are outliers and False
        otherwise.

        Args:

            points (np.ndarray[float]) - 1-D array of observations

            threshold (float) - Maximum modified z-score. Observations with a modified z-score (based on the median absolute deviation) greater are classified as outliers.

        Returns:

            mask (np.ndarray[bool])

        References:

            Boris Iglewicz and David Hoaglin (1993), "Volume 16: How to Detect and Handle Outliers", The ASQC Basic References in Quality Control:
            Statistical Techniques, Edward F. Mykytka, Ph.D., Editor.

        """
        if len(points.shape) == 1:
            points = points[:,None]
        median = np.median(points, axis=0)
        diff = np.sum((points - median)**2, axis=-1)
        diff = np.sqrt(diff)
        med_abs_deviation = np.median(diff)

        modified_z_score = 0.6745 * diff / med_abs_deviation

        # exclude lower bound
        modified_z_score[points.ravel()<median] = 0

        return modified_z_score > threshold

    @property
    def edge_angles(self):
        """ Angular distance of each edge about origin. """
        angles = np.abs(np.diff(self.angles[self.edge_list], axis=1)).ravel()
        angles[angles>np.pi] = 2*np.pi-angles[angles>np.pi]
        return angles

    @property
    def edge_radii(self):
        """ Minimum node radius in each edge. """
        return self.radii[self.edge_list].min(axis=1)

    @property
    def angle_threshold(self):
        """ Predicted upper bound on edge angles. """
        num_sigma = 1.
        return (3+2.5*num_sigma) * (self.size**(-0.5))

    def filter_by_angle(self):

        # exclude outer edges that span too wide an angle
        edge_radii = self.edge_radii
        edge_angles = self.edge_angles
        outer_edge_mask = edge_radii > np.percentile(edge_radii, 50)
        wide_angle_mask = edge_angles >= self.angle_threshold
        excluded_edge_mask = np.logical_and(outer_edge_mask, wide_angle_mask)

        # determine accepted/rejected edges
        rejected = self.edge_list[excluded_edge_mask]
        accepted = self.edge_list[~excluded_edge_mask]

        # find disconnected nodes
        disconnected = self.find_disconnected_nodes(self.nodes, accepted)

        # sort rejected edges by length
        lengths = self.edge_lengths[excluded_edge_mask]
        sort_indices = np.argsort(lengths)
        rejected = rejected[sort_indices]

        # add shortest edge for each disconnected node
        if disconnected.size > 0:
            f = np.vectorize(lambda node: self.find_first_edge(rejected, node))
            connecting = rejected[f(disconnected)]
            accepted = np.vstack((accepted, connecting, connecting[::-1]))

        return accepted



import numpy as np


class Defaults:
    """
    Object contains default parameters. Defaults are stored as dictionaries within the instance attributes:

    Attributes:

        parameters (dict) - {parameter set: parameters dict} pairs for:

        preprocessing (dict) - default parameters for image preprocessing

        seeds (dict) - default parameters for seed detection

        segmentation (dict) - default parameters for segmentation

    """

    def __init__(self):
        """
        Instantiate object with default parameter dictionaries as attributes
        """

        self.parameters = {

            'preprocessing':
                {
                'median_radius': 2,
                'gaussian_sigma': (2, 2),
                'clip_limit': 0.03,
                'clip_factor': 20
                },

            'seeds':
                {
                'sigma': 2,
                'min_distance': 1,
                'num_peaks': np.inf
                },

            'segmentation':
                {
                'sigma': 0.5,
                'watershed_line': True
                }}

    def __call__(self, key, specified):
        """
        Adds default values for unspecified <key> parameters.

        Args:

            key (str) - name of parameter set

            specified (dict) - {name: value} pairs for specified <key> parameters

        Returns:

            parameters (dict) - {name: value} pairs for all <key> parameters

        """
        return self.append_defaults(key, specified)

    def append_defaults(self, key, specified):
        """
        Add default values for unspecified <key> parameters.

        Args:

            key (str) - name of parameter set

            specified (dict) - {name: value} pairs for specified <key> parameters

        Returns:

            parameters (dict) - {name: value} pairs for all <key> parameters

        """
        assert key in self.parameters.keys(), 'Parameter set not recognized.'
        default = self.parameters[key]
        return self._append_defaults(specified, default)

    @staticmethod
    def _append_defaults(specified, default):
        """
        Add default values for unspecified parameters.

        Args:

            specified (dict) - {name: value} pairs for specified parameters

            default (dict) - {name: default value} pairs for all parameters

        Returns:

            parameters (dict) - {name: value} pairs for all parameters

        """
        parameters = {}
        for k, v in default.items():
            if k in specified.keys():
                parameters[k] = specified[k]
            else:
                parameters[k] = v
        return parameters



from time import time
from os.path import curdir
from clones.validation.arguments import RunArguments
from clones.validation.batch import BatchBenchmark


# ======================== PARSE SCRIPT ARGUMENTS =============================

args = RunArguments(description='Batch benchmark arguments.')
job_path = args['path']
train = args['train_globally']

# ============================= RUN SCRIPT ====================================

start_time = time()


# run each simulation in job file
with open(job_path, 'r') as job_file:

    # run each simulation
    for path in job_file.readlines():

        path = path.strip()

        # load benchmark
        benchmark = BatchBenchmark.load(path)
        benchmark.batch.root = curdir

        # evaluate benchmark
        benchmark.run(train=train)

        # save benchmark
        benchmark.save(path)

# print runtime to standard out
runtime = time() - start_time
print('BATCH COMPLETED IN {:0.2f}.\n'.format(runtime))



from copy import deepcopy
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize, ListedColormap
from matplotlib.tri import Triangulation
import networkx as nx
from collections import Counter

from .triangulation import LocalTriangulation
from .infomap import InfoMap
from .correlation import SpatialCorrelation


class TopologicalProperties:
    """ Topological properties for Graph objects. """

    @property
    def num_nodes(self):
        """ Number of nodes. """
        return len(self.data)

    @property
    def nodes(self):
        """ Unique nodes in graph. """
        return np.array(sorted(np.unique(self.edges)), dtype=int)

    @property
    def nodes_order(self):
        """ Indices that sort nodes by positional index in <self.data>. """
        return np.argsort(self.position_map(self.nodes))

    @property
    def edges(self):
        """ Distance-filtered edges. """
        return self.node_map(self.tri.edges)

    @property
    def edge_list(self):
        """ Distance-filtered edges as (from, to) tuples. """
        return [(int(e[0]), int(e[1]), None) for e in self.edges]

    @property
    def adjacency(self):
        """ Adjacency matrix ordered by <self.nodes>. """
        return nx.to_numpy_array(self.G, nodelist=self.nodes)

    @property
    def adjacency_positional(self):
        """ Adjacency matrix ordered by positional index in <self.data>. """
        return self.adjacency[self.nodes_order, :][:, self.nodes_order]


class SpatialProperties:
    """ Spatial properties for Graph objects. """

    @property
    def node_positions(self):
        """ Assign 2D coordinate positions to nodes. """
        node_positions = {}
        for k in self.nodes:
            i = self.position_map(k)
            node_positions[k] = np.array([self.tri.x[i], self.tri.y[i]])
        return node_positions

    @property
    def node_positions_arr(self):
        """ N x 2 array of node coordinates, ordered by positional index."""
        return self.data[self.xykey].values

    @property
    def edge_lengths(self):
        """ Unique edge lengths. """
        return sorted(self.distance_matrix[self.adjacency.nonzero()])[::2]

    @property
    def median_edge_length(self):
        """ Median edge length. """
        return np.median(self.edge_lengths)

    @staticmethod
    def _distance_matrix(xy):
        """ Returns euclidean distance matrix between all points in <xy>. """

        x, y = xy.T
        x = x.reshape(-1, 1)
        y = y.reshape(-1, 1)
        x_component = np.repeat(x**2, x.size, axis=1) + np.repeat(x.T**2, x.size, axis=0) - 2*np.dot(x, x.T)
        y_component = np.repeat(y**2, y.size, axis=1) + np.repeat(y.T**2, y.size, axis=0) - 2*np.dot(y, y.T)

        return np.sqrt(x_component + y_component)

    @property
    def distance_matrix(self):
        """ Euclidean distance matrix between all nodes. """
        return self._distance_matrix(self.node_positions_arr)

    @staticmethod
    def get_matrix_upper(matrix):
        """
        Return upper triangular portion of a 2-D matrix.

        Parameters:

            matrix (2D np.ndarray)

        Returns:

            upper (1D np.ndarray) - upper triangle, ordered row then column

        """
        return matrix[np.triu_indices(len(matrix), k=1)]

    @property
    def unique_distances(self):
        """ Upper triangular portion of euclidean distance matrix. """
        return self.get_matrix_upper(self.distance_matrix)

    @staticmethod
    def evaluate_fluctuations(values):
        """
        Construct pairwise fluctuation matrix for <values>.

        Args:

            values (1D np.ndarray[float]) - attribute values

        Returns:

            fluctuations (2D np.ndarray[float]) - pairwise fluctuations

        """

        # compute correlation matrix and return upper triangular portion
        v = (values - values.mean())
        fluctuations = np.dot(v, v.T) / np.var(values)
        return fluctuations

    def get_fluctuations_matrix(self, attribute, log=True):
        """
        Returns normalized pairwise fluctuations of <attribute> value for each node in the graph.

        Args:

            attribute (str) - name of attribute

            log (bool) - if True, log-transform attribute values

        Returns:

            fluctuations (2D np.ndarray[float]) - pairwise fluctuations

        """
        attribute_values = self.data[attribute].values.reshape(-1, 1)
        if log:
            attribute_values = np.log(attribute_values)
        return self.evaluate_fluctuations(attribute_values)

    def _mark_boundary(self, attribute='genotype', max_edges=0):
        """
        Mark boundaries between nodes with differing attribute values. Boundaries are identified by nodes that share an edge with another node whose attribute value differs.

        Args:

            attribute (str) - used to identify distinct groups of nodes

            max_edges (int) - max number of transgroup edges for interior nodes

        Returns:

            interior (np.vector[bool]) - mask for interior points

        """

        # assign genotype to edges
        assign_genotype = np.vectorize(dict(self.data[attribute]).get)
        edge_genotypes = assign_genotype(self.edges)

        # find edges traversing clones
        boundaries = (edge_genotypes[:, 0] != edge_genotypes[:, 1])

        # get number of clone-traversing edges per node
        boundary_edges = self.edges[boundaries]
        edge_counts = Counter(boundary_edges.ravel())

        # assign boundary label to nodes with too many clone-traversing edges
        boundary_nodes = [n for n, c in edge_counts.items() if c>max_edges]
        interior = np.ones(self.num_nodes, dtype=bool)
        interior[self.position_map(boundary_nodes)] = False

        return interior


class GraphVisualizationMethods:
    """ Methods for visualizing a Graph instance. """

    def plot_edges(self, ax=None, **kwargs):
        """
        Plot triangulation edges.

        Args:

            ax (matplotlib.axes.AxesSubplot)

            kwargs: keyword arguments for matplotlib.pyplot.triplot

        """
        if ax is None:
            fig, ax = plt.subplots()
        lines, markers = plt.triplot(self.tri, **kwargs)

    def label_triangles(self, label_by='genotype'):
        """
        Label each triangle with most common node attribute value.

        Args:

            label_by (str) - node attribute used to label each triangle

        Returns:

            labels (np.ndarray[int]) - labels for each triangle

        """

        # get triangulation vertices
        vertices = self.node_map(self.tri.triangles)

        # get value of each node
        get_level = lambda node_id: self.data[label_by].loc[node_id]
        levels = np.apply_along_axis(get_level, axis=1, arr=vertices)

        # aggregate within triangles
        get_mode = lambda x: Counter(x).most_common(1)[0][0]
        labels = np.apply_along_axis(get_mode, axis=1, arr=levels)

        return labels

    def plot_triangles(self,
                       label_by='genotype',
                       cmap=None,
                       ax=None,
                       **kwargs):
        """
        Plot triangle faces using tripcolor.

        Args:

            label_by (str) - data attribute used to color each triangle

            cmap (matplotlib.colors.ColorMap) - colormap for attribute values

            ax (matplotlib.axes.AxesSubplot)

            kwargs: keyword arguments for plt.tripcolor

        """
        if ax is None:
            fig, ax = plt.subplots()

        # assign triangle colors
        colors = self.label_triangles(label_by=label_by)

        # define colormap
        if cmap is None:
            cmap = ListedColormap(['y', 'm', 'k'], N=3)

        # plot triangle faces
        ax.tripcolor(self.tri,
                     facecolors=colors,
                     cmap=cmap,
                     lw=0,
                     edgecolor='none',
                     antialiased=True,
                     **kwargs)

    def show(self, ax=None, colorby=None, disconnect=False, **kwargs):
        """
        Visualize graph.

        Args:

            ax (matplotlib.axes.AxesSubplot) - if None, create figure

            colorby (str) - node attribute used to assign node/edge colors

            disconnect (bool) - if True, remove edges between nodes whose colorby values differ

            kwargs: keyword arguments for NetworkxGraphVisualization.draw

        """
        if colorby is not None:
            msg = 'Colorby attribute must be an integer type.'
            assert self.data[colorby].dtype in (np.integer, int), msg

        # construct graph
        if colorby is not None:
            G = self.get_networkx(colorby)
        else:
            G = self.G

        # disconnect nodes not sharing the <colorby> attribute
        if disconnect:
            is_different = lambda u,v: G.node[u][colorby] != G.node[v][colorby]
            removed_edges = [edge for edge in G.edges if is_different(*edge)]
            G.remove_edges_from(removed_edges)

        # draw graph
        vis = NetworkxGraphVisualization(G, self.node_positions)
        vis.draw(ax=ax, colorby=colorby, **kwargs)


class CommunityDetection:
    """ Methods for detecting communities in a Graph. """

    def detect_communities(self, **kwargs):
        """
        Detect communities using InfoMap clustering.

        Accepts keyword arguments for InfoMap, including:

            twolevel (bool) - if True, perform two-level clustering, otherwise defaults to multi-level clustering

            N (int) - number of trials

        """
        self.imap = InfoMap(self.edge_list, **kwargs)

    def _assign_community(self, level=None):
        """
        Assign communities using InfoMap clustering.

        Args:

            level (int) - module level at which aggregation occurs, starting from the finest resolution

        Returns:

            labels (np.ndarray[int]) - community labels ordered by positional index

        """
        return self.imap(self.nodes[self.nodes_order], level)

    def assign_community(self, level=None, key='community'):
        """
        Assign communities using InfoMap clustering.

        Args:

            level (int) - module level at which aggregation occurs, starting from the finest resolution

            key (str) - name of community attribute

        """
        self.data[key] = self._assign_community(level=level)


class Graph(TopologicalProperties,
            SpatialProperties,
            CommunityDetection,
            GraphVisualizationMethods):
    """
    Object provides an undirected unweighted graph connecting adjacent cells.

    Attributes:

        data (pd.DataFrame) - cell measurement data (nodes)

        xykey (list) - attribute keys for node x/y positions

        G (nx.Graph) - undirected graph instance

        nodes (np.ndarray[int]) - node indices

        edges (np.ndarray[int]) - pairs of connected node indices

        node_map (vectorized func) - maps positional index to node index

        position_map (vectorized func) - maps node index to positional index

        tri (matplotlib.tri.Triangulation) - triangulation of node positions

    """

    def __init__(self, data, xykey=None):
        """
        Instantiate Graph object.

        Args:

            data (pd.DataFrame) - cell measurement data (nodes)

            xykey (list) - attribute keys for node x/y positions

        """

        # set xykey
        if xykey is None:
            xykey = ['centroid_x', 'centroid_y']
        self.xykey = xykey

        # store data
        self.data = data

        # instantiate community detection
        self.community_labels = None
        self.imap = None

        # define mapping from position to node index
        position_to_node = dict(enumerate(data.index))
        self.node_map = np.vectorize(position_to_node.get)

        # define reverse map
        node_to_position = {v: k for k, v in position_to_node.items()}
        self.position_map = np.vectorize(node_to_position.get)

        # triangulate
        self.tri = self._construct_triangulation(data, xykey)

        # build networkx graph instance
        self.G = self.get_networkx()

    def copy(self):
        """ Returns deep copy of graph instance. """
        return deepcopy(self)

    def get_subgraph(self, ind):
        """ Instantiate subgraph from DataFrame indices. """
        return Graph(self.data.loc[ind], xykey=self.xykey)

    def get_networkx(self, *node_attributes):
        """
        Returns networkx instance of graph.

        Args:

            node_attributes (str) - attributes to be added for each node

        """
        G = nx.Graph()
        G.add_weighted_edges_from(self.edge_list)

        # add node attributes
        for attr in node_attributes:
            if attr is not None:
                values_dict = dict(self.data.loc[self.nodes][attr])
                nx.set_node_attributes(G, name=attr, values=values_dict)

        return G

    def get_correlations(self, attribute, log=True):
        """
        Returns SpatialCorrelation object for <attribute>.

        Args:

            attribute (str) - name of attribute

            log (bool) - if True, log-transform attribute values

        Returns:

            correlations (SpatialCorrelation)

        """
        d_ij = self.unique_distances
        C = self.get_fluctuations_matrix(attribute, log=log)
        C_ij = self.get_matrix_upper(C)
        return SpatialCorrelation(d_ij, C_ij)

    @staticmethod
    def _construct_triangulation(data, xykey, **kwargs):
        """
        Construct Delaunay triangulation with edge filter.

        Args:

            data (pd.DataFrame) - node measurement data

            xykey (list) - attribute keys for node x/y coordinates

            kwargs: keyword arguments for triangulation

        """
        pts = data[xykey].values
        return LocalTriangulation(*pts.T, **kwargs)


class WeightedGraph(Graph):
    """
    Object provides an undirected weighted graph connecting adjacent cells. Edge weights are evaluated based on the similarity of expression between pairs of connected nodes. Node similariy is based on the cell measurement data attribute specified by the 'weighted_by' parameter.

    Attributes:

        weighted_by (str) - data attribute used to weight edges

        imap (spatial.InfoMap) - community detection

        community_labels (np.ndarray[int]) - community label for each node

        logratio (bool) - if True, weight edges by log ratio

        distance (bool) - if True, weights edges by distance rather than similarity

    Inherited attributes:

        data (pd.DataFrame) - cell measurement data (nodes)

        xykey (list) - attribute keys for node x/y positions

        nodes (np.ndarray[int]) - node indices

        edges (np.ndarray[int]) - pairs of connected node indices

        node_map (vectorized func) - maps positional index to node index

        position_map (vectorized func) - maps node index to positional index

        tri (matplotlib.tri.Triangulation) - triangulation of node positions

    """

    def __init__(self, data, weighted_by,
                 xykey=None,
                 logratio=True,
                 distance=False):
        """
        Instantiate weighted graph.

        Args:

            data (pd.DataFrame) - cell measurement data

            weighted_by (str) - data attribute used to weight edges

            xykey (list) - attribute keys for node x/y positions

            logratio (bool) - if True, weight edges by log ratio

            distance (bool) - if True, weights edges by distance

        """

        # set attributes
        self.weighted_by = weighted_by
        self.logratio = logratio
        self.distance = distance

        # instantiate graph
        super().__init__(data, xykey)

    @property
    def edge_list(self):
        """ Distance-filtered edges as (from, to, weight) tuples. """
        if self.weighted_by not in (None, 'none', 'None'):
            weights = self.evaluate_edge_weights()
            unpack = lambda e, w: (int(e[0]), int(e[1]), w)
            edge_list = [unpack(e, w) for e, w in zip(self.edges, weights)]
        else:
            edge_list = super().edge_list

        return edge_list

    def evaluate_edge_weights(self):
        """
        Evaluate edge weights.

        Returns:

            weights (np.ndarray[float]) - edge weights

        """
        wf = WeightFunction(self.data,
                            weighted_by=self.weighted_by,
                            distance=self.distance)
        return wf.assess_weights(self.edges, logratio=self.logratio)


class WeightFunction:
    """
    Object for weighting graph edges by similarity.

    Attributes:

        data (pd.DataFrame) - nodes data

        weighted_by (str) - node attribute used to assess similarity

        values (pd.Series) - node attribute values

        distance (bool) - if True, weights edges by distance

    """

    def __init__(self, data, weighted_by='r', distance=False):
        """
        Instantiate edge weighting function.

        Args:

            data (pd.DataFrame) - nodes data

            weighted_by (str) - node attribute used to assess similarity

            distance (bool) - if True, weights edges by distance

        """
        self.data = data
        self.weighted_by = weighted_by
        self.values = data[weighted_by]
        self.distance = distance

    def difference(self, i, j):
        """
        Evaluate difference in values between nodes i and j.

        Args:

            i, j (ind) - node indices

        Returns:

            difference (float)

        """
        return np.abs(self.values.loc[i] - self.values.loc[j])

    def logratio(self, i, j):
        """
        Evaluate log ratio between nodes i and j.

        Args:

            i, j (ind) - node indices

        Returns:

            logratio (float)

        """
        return np.abs(np.log(self.values.loc[i]/self.values.loc[j]))

    def assess_weights(self, edges, logratio=False):
        """
        Evaluate edge weights normalized by mean difference in node values.

        Args:

            edges (list of (i, j) tuples) - edges between nodes i and j

            logratio (bool) - if True, weight edges by logratio

        Returns:

            weights (np.ndarray[float]) - edge weights

        """
        if logratio:
            energy = np.array([self.logratio(*e) for e in edges])
        else:
            energy = np.array([self.difference(*e) for e in edges])
        weights = np.exp(-energy/np.mean(energy))

        # invert similarities to distances
        if self.distance:
            weights = 1 - weights

        return weights


class NetworkxGraphVisualization:
    """
    Object for visualizing a NetworkX Graph object.

    Attributes:

        G (nx.Graph) - networkx graph object

        pos (np.ndarray[float]) - 2D node positions

    """

    def __init__(self, G, pos):
        """
        Instantiate graph visualization.

        Args:

            G (nx.Graph) - networkx graph object

            pos (np.ndarray[float]) - 2D node positions

        """
        self.G = G
        self.pos = pos

    def build_cmap(self, colorby):
        """ Build colormap. """
        levels = [v for k,v in nx.get_node_attributes(self.G, colorby).items()]
        N = np.unique(levels).size
        colors = np.random.random(size=(N, 3))
        return ListedColormap(colors, 'indexed', N)

    def draw(self,
             ax=None,
             colorby=None,
             edge_color='k',
             node_color='k',
             cmap=None,
             **kwargs):
        """
        Draw graph.

        Args:

            ax (matplotlib.axes.AxesSubplot) - axis on which to draw graph

            colorby (str) - node attribute on which nodes/edges are colored

            edge_color, node_color (str) - edge/node colors, overrides colorby

            node_cmap (matplotlib.colors.ColorMap) - node colormap

        """

        assert (colorby is not None or node_color is not None), 'Either node color or colorby attribute must be specified.'
        assert (colorby is not None or edge_color is not None), 'Either edge color or colorby attribute must be specified.'

        # create figure
        if ax is None:
            fig, ax = plt.subplots(figsize=(10, 10))

        # build colormap
        if cmap is None and None in (edge_color, node_color):
            cmap = self.build_cmap(colorby=colorby)

        # assign node color
        if node_color is None:
            get_color = lambda x: cmap(int(x))
            node_colors = [get_color(v) for k, v in nx.get_node_attributes(self.G, colorby).items()]
        else:
            node_colors = [node_color for _ in self.G.nodes]

        # assign edge color
        if edge_color is None:
            node_value = lambda index: self.G.node[index][colorby]
            node_color = lambda index: cmap(node_value(index))
            edge_colors = [node_color(u) for u,v in self.G.edges]
        else:
            edge_colors = [edge_color for _ in self.G.edges]

        # draw graph
        self._draw(ax, self.G, self.pos, node_colors, edge_colors, **kwargs)

    @staticmethod
    def _draw(ax,
              G,
              pos,
              node_colors,
              edge_colors,
              node_alpha=1,
              node_size=20,
              node_edgewidth=0.,
              lw=3,
              edge_alpha=0.5,
              **kwargs):
        """
        Draw graph.

        Args:

            G (nx.Graph) - graph object

            pos (np.ndarray[float]) - node xy positions in space (graph layout)

            node_colors (array like) - node values

            edge_colors (array like) - edge colors

            node_alpha (float) - node transparency

            node_size (float) - node size

            node_edgewidth (float) - linewidth of node outline

            lw (float) - maximum edge linewidth

            edge_alpha (float) - edge line transparency

            kwargs: keyword arguments for nx.draw_networkx_nodes

        """

        # extract edge weights
        edge_weights = [x[-1]['weight'] for x in G.edges(data=True)]

        # draw edges
        norm = Normalize(vmin=min(edge_weights), vmax=max(edge_weights))
        edge_widths = [norm(w)*lw for w in edge_weights]
        nx.draw_networkx_edges(G, pos,
                               ax=ax,
                               edge_color=edge_colors,
                               width=edge_widths,
                               alpha=edge_alpha)

        # draw nodes
        nodeCollection = nx.draw_networkx_nodes(G,
            pos=pos,
            ax=ax,
            node_color=node_colors,
            node_alpha=node_alpha,
            node_size=node_size,
            linewidths=node_edgewidth,
            **kwargs)

        ax.axis('off')



from os.path import join, exists
from os import mkdir
from copy import deepcopy
import numpy as np
import networkx as nx
from collections import Counter

from ..utilities import IO

from .spatial.sampling import NeighborSampler, CommunitySampler, RadialSampler
from .classification.mixtures import BivariateMixtureClassifier
from .model_selection import BivariateModelSelection


class AnnotationIO:
    """
    Methods for saving and loading an Annotation instance.
    """

    @property
    def parameters(self):
        """ Dictionary of parameter values. """
        return {k:v for k,v in self.__dict__.items() if k != 'classifier'}

    def save(self, dirpath, data=False, image=False, **kwargs):
        """
        Save annotator to specified path.

        Args:

            dirpath (str) - directory in which annotator is to be saved

            data (bool) - if True, save training data

            image (bool) - if True, save classifier image

            kwargs: keyword arguments for image rendering

        """

        # save parameters
        io = IO()
        io.write_json(join(dirpath, 'annotation.json'), self.parameters)

        # save classifier
        if self.classifier is not None:
            self.classifier.save(dirpath, data=data, image=image, **kwargs)

    @classmethod
    def load(cls, path):
        """
        Load annotator from file.

        Args:

            path (str) - path to annotation directory

        Returns:

            annotator (Annotation derivative)

        """

        io = IO()

        # load parameters
        parameters = io.read_json(join(path, 'annotation.json'))
        attribute = parameters.pop('attribute')

        # instantiate annotator
        annotator = cls(attribute, **parameters)

        # load classifier
        classifier_path = join(path, 'classifier')
        if exists(classifier_path):
            annotator.classifier = BivariateMixtureClassifier.load(classifier_path)

        return annotator


class Annotation(AnnotationIO):
    """
    Object for assigning labels to measurements. Object is trained on one or more graphs by fitting a bivariate mixture model and using a model selection procedure to select an optimal number of components.

    The trained model may then be used to label measurements in other graphs, either through direct prediction via the bivariate mixture model or through a hybrid prediction combining the bivariate mixture model with a marginal univariate model.

    Attributes:

        classifier (Classifier derivative) - callable object

        attribute (str) - attribute used to determine labels

        sampler_type (str) - either 'radial', 'neighbors', 'community'

        sampler_kwargs (dict) - keyword arguments for sampler

        min_num_components (int) - minimum number of mixture components

        max_num_components (int) - maximum number of mixture components

        num_labels (int) - maximum number of unique labels to be assigned

    Parameters:

        kwargs: keyword arguments for Classifier

    """

    def __init__(self, attribute,
                 sampler_type='radial',
                 sampler_kwargs={},
                 min_num_components=3,
                 max_num_components=10,
                 num_labels=3):
        """
        Instantiate annotation object.

        Args:

            attribute (str) - name of attribute used to classify cells

            sampler_type (str) - either 'radial', 'neighbors', 'community'

            sampler_kwargs (dict) - keyword arguments for sampler

            min_num_components (int) - minimum number of mixture components

            max_num_components (int) - maximum number of mixture components

            num_labels (int) - maximum number of unique labels to be assigned

        """
        self.attribute = attribute
        self.sampler_type = sampler_type
        self.sampler_kwargs = sampler_kwargs
        self.min_num_components = min_num_components
        self.max_num_components = max_num_components
        self.num_labels = num_labels

    def __call__(self, graph, **kwargs):
        """ Returns labels for a graph of measurements. """
        return self.annotate(graph, **kwargs)

    @classmethod
    def from_data(cls, data, attribute, xykey=None, **kwargs):
        """
        Instantiate annotation object from measurement data.

        Args:

            data (pd.DataFrame) - measurement data containing <attribute>, as well as <xykey> fields

            attribute (str) - name of attribute used to classify cells

            xykey (list) - name of attributes defining measurement x/y position

            kwargs: keyword arguments for Annotation

        Returns:

            annotator (Annotation derivative)

        """
        annotator = cls(attribute, **kwargs)
        annotator.train(WeightedGraph(data, attribute, xykey=xykey))
        return annotator

    @classmethod
    def from_layer(cls, layer, attribute, **kwargs):
        """
        Instantiate from layer.

        Args:

            layer (data.Layer) - image layer instance

            attribute (str) - name of attribute used to classify cells

            kwargs: keyword arguments for Annotation

        Returns:

            annotator (Annotation derivative)

        """
        annotator = cls(attribute, **kwargs)
        annotator.train(layer.graph)
        return annotator

    @classmethod
    def copy(cls, src):
        """ Instantiate from another <source> annotator instance. """
        dst = cls(src.attribute)
        dst.__dict__.update(src.__dict__)
        return dst

    def train(self, *graphs):
        """
        Train classifier on a series of graphs.

        Args:

            graphs (Graph or WeightedGraph) - graphs of adjacent measurements

        """

        # generate sample
        if self.sampler_type == 'radial':
            data, keys = RadialSampler.multisample(self.attribute,
                                                *graphs,
                                                **self.sampler_kwargs)

        elif self.sampler_type == 'community':
            data, keys = CommunitySampler.multisample(self.attribute,
                                                *graphs,
                                                **self.sampler_kwargs)

        elif self.sampler_type == 'neighbors':
            data, keys = NeighborSampler.multisample(self.attribute,
                                               *graphs,
                                               **self.sampler_kwargs)

        else:
            raise ValueError('Sampler type ''{:s}'' not recognized.'.format(self.sampler_type))

        # run model selection
        selector = BivariateModelSelection(data,
            keys,
            min_num_components=self.min_num_components,
            max_num_components=self.max_num_components,
            num_labels=self.num_labels)

        # store BIC-optimal model
        self.classifier = selector.BIC_optimal

        return selector

    def get_sampler(self, graph, sampler_type=None, sampler_kwargs=None):
        """
        Instantiate sampler.

        Args:

            graph (spatial.WeightedGraph)

            sampler_type (str) - either 'radial', 'neighbors' or 'community'

            sampler_kwargs (dict) - keyword arguments for sampling

        Returns:

            sampler

        """

        # use default sampler configuration if none is specified
        if sampler_type is None:
            sampler_type = self.sampler_type

        if sampler_kwargs is None:
            sampler_kwargs = self.sampler_kwargs

        # instantiate sampler
        if sampler_type == 'radial':
            sampler = RadialSampler(graph, self.attribute, **sampler_kwargs)
        elif sampler_type == 'community':
            sampler = CommunitySampler(graph, self.attribute, **sampler_kwargs)
        elif sampler_type == 'neighbors':
            sampler = NeighborSampler(graph, self.attribute, **sampler_kwargs)
        else:
            raise ValueError('Sampler type ''{:s}'' not recognized.'.format(sampler_type))

        return sampler

    def get_sample(self, graph, sampler_type, sampler_kwargs):
        """
        Get sample to be annotated. A sample consists of a columns of measured levels adjoined to a column of levels averaged over the neighborhood of each measurement.

        Args:

            graph (spatial.WeightedGraph)

            sampler_type (str) - either 'radial', 'neighbors' or 'community'

            sampler_kwargs (dict) - keyword arguments for sampling

        Returns:

            sample (np.ndarray[float]) - sampled levels

        """

        # instantiate sampler
        sampler = self.get_sampler(graph, sampler_type, sampler_kwargs)

        # generate sample
        sample = sampler.sample

        # log transform sample
        if self.classifier.log:
            sample = np.log(sample)

        return sample

    def combine_posteriors(self, posterior, marginal_posterior, threshold=0.8):
        """
        Replace uncertain posterior probablilities with their more certain marginal counterparts. If the maximum marginal posterior probability for a given sample does not meet the specified threshold while the maximum bivarite posterior probability does, the latter value is used. Otherwise, the marginal value is used.

        Args:

            posterior (np.ndarray[float]) - posterior probabilities of each label

            marginal_posterior (np.ndarray[float]) - marginal posterior probabilities of each label

            threshold (float) - minimum marginal posterior probability of a given label before spatial context is considered

        Returns:

            combined (np.ndarray[float])

        """
        combined = deepcopy(marginal_posterior)
        mask = np.logical_and(posterior.max(axis=1) > threshold,
                              marginal_posterior.max(axis=1) < threshold)
        combined[mask, :] = posterior[mask, :]
        return combined

    def evaluate_marginal_posterior(self, sample, margin):
        """
        Evaluates posterior probability of each label using only the specified marginal distribution.

        Args:

            sample (np.ndarray[float]) - sample values

            margin (int) - index of desired margin

        Returns:

            marginal_posterior (np.ndarray[float])

        """
        return self.classifier[margin].posterior(sample[:, [margin]])

    @staticmethod
    def diffuse_posteriors(graph, posterior, alpha=0.9):
        """
        Diffuse estimated posterior probabilities of each label along the weighted edges of the graph.

        Args:

            graph (Graph) - graph connecting adjacent measurements

            posterior (np.ndarray[float]) - posterior probabiltiy of each label

            alpha (float) - attenuation factor

        Returns:

            diffused_posteriors (np.ndarray[float])

        """

        # compile normalized adjancy matrix
        adjacency = graph.adjacency_positional
        adjacency /= adjacency.sum(axis=0)

        # evaluate centrality
        external = np.eye(*adjacency.shape) - (alpha * adjacency)
        internal = (1 - alpha) * posterior
        diffused_posteriors = np.linalg.solve(external, internal)

        return diffused_posteriors

    def annotate(self, graph,
                 bivariate_only=False,
                 threshold=0.8,
                 alpha=0.9,
                 sampler_type=None,
                 sampler_kwargs=None):
        """
        Annotate graph of measurements.

        Args:

            graph (spatial.WeightedGraph)

            bivariate_only (bool) - if True, only use posteriors evaluated using the bivariate mixture model. Otherwise, use the marginal univariate posterior by default, replacing uncertain values with their counterparts estimated by the bivariate model.

            threshold (float) - minimum marginal posterior probability of a given label before spatial context is considered

            alpha (float) - attenuation factor

            sampler_type (str) - either 'radial', 'neighbors' or 'community'

            sampler_kwargs (dict) - keyword arguments for sampling

        Returns:

            labels (np.ndarray[int]) - labels for each measurement in graph

        """

        # get sample data
        sample = self.get_sample(graph=graph,
                                 sampler_type=sampler_type,
                                 sampler_kwargs=sampler_kwargs)

        # classify sample
        posterior = self.classifier.posterior(sample)

        # combine with posteriors estimated by univariate marginal classifier
        if not bivariate_only:
            marginal = self.evaluate_marginal_posterior(sample, 0)
            posterior = self.combine_posteriors(posterior, marginal, threshold)

        # diffuse posteriors
        if alpha is not None:
            posterior = self.diffuse_posteriors(graph, posterior, alpha=alpha)

        # assign labels
        labels = posterior.argmax(axis=1)

        return labels



from time import time
from clones.validation.arguments import SweepArguments
from clones.validation.sweep import SweepBenchmark


# ======================== PARSE SCRIPT ARGUMENTS =============================

args = SweepArguments(description='SweepBenchmark arguments.')

# ============================= RUN SCRIPT ====================================

start = time()
sweep_benchmark = SweepBenchmark.load(args['path'])
sweep_benchmark.aggregate()
runtime = time() - start

print('AGGREGATION COMPLETED IN {:0.2f} seconds.'.format(runtime))



from os.path import join, abspath, isdir
from glob import glob
import numpy as np
import pandas as pd

from ..utilities import UserPrompts

from .stacks import Stack


class Experiment:
    """
    Object represents a collection of 3D RGB image stacks collected under the same experimental conditions.

    Attributes:

        path (str) - path to experiment directory

        _id (str) - name of experiment

        stack_ids (list of str) - unique stack ids within experiment

        stack_dirs (dict) - {stack_id: stack_directory} tuples

        count (int) - counter for stack iteration

    """

    def __init__(self, path):
        """
        Instantiate experiment object.

        Args:

            path (str) - directory with subdirectories of 3D RGB image stacks

        """

        # set path to experiment directory
        self.path = abspath(path)

        # set experiment ID
        self._id = path.split('/')[-1]

        # set stack paths
        stack_paths = [p for p in glob(join(self.path, '*')) if isdir(p)]
        get_stack_id = lambda x: x.rsplit('/', maxsplit=1)[-1]
        self.stack_dirs = {get_stack_id(p): p for p in stack_paths}
        self.stack_ids = sorted(self.stack_dirs.keys())

        # reset stack iterator count
        self.count = 0

        # check if stacks have been initialized, if not prompt user
        if not self.is_initialized:
            self.prompt_initialization()

    def __getitem__(self, stack_id):
        """ Load stack. """
        return self.load_stack(stack_id, full=False)

    def __iter__(self):
        """ Iterate across stacks. """
        self.count = 0
        return self

    def __next__(self):
        """ Return next stack. """
        if self.count < len(self.stack_ids):
            stack_id = self.stack_ids[self.count]
            stack = self.__getitem__(stack_id)
            self.count += 1
            return stack
        else:
            raise StopIteration

    @property
    def is_initialized(self):
        """ Returns True if Experiment has been initialized. """
        for stack_dir in self.stack_dirs.values():
            if not Stack._check_if_initialized(stack_dir):
                return False
        return True

    def prompt_initialization(self):
        """ Ask user whether to initialize all stack directories. """
        msg = 'Incomplete stack directories found. Initialize them?'
        user_response = UserPrompts.boolean_prompt(msg)
        if user_response:
            msg = 'Please enter an image bit depth:'
            bit_depth = UserPrompts.integer_prompt(msg)
            if bit_depth is not None:
                self.initialize(bit_depth)
            else:
                raise ValueError('User response not recognized, stacks have not been initialized.')

    def initialize(self, bit_depth):
        """
        Initialize a collection of image stacks.

        Args:

            bit_depth (int) - bit depth of raw tif (e.g. 12 or 16). Value will be read from the stack metadata if None is provided. An error is raised if no value is found.

        """
        for stack_id in self.stack_ids:
            _ = self.load_stack(stack_id, full=False, bit_depth=bit_depth)

    def load_stack(self, stack_id, full=False, **kwargs):
        """
        Load 3D RGB image stack.

        Args:

            stack_id (str or int) - desired stack

            full (bool) - if True, load full 3D image from tif file

        Returns:

            stack (Stack)

        """
        stack = Stack(self.stack_dirs[str(stack_id)], **kwargs)
        if full:
            stack.load_image()
        return stack

    def aggregate_measurements(self,
                               selected_only=False,
                               exclude_boundary=False,
                               raw=False,
                               use_cache=True):
        """
        Aggregate measurements from each stack.

        Args:

            selected_only (bool) - if True, exclude cells outside the ROI

            exclude_boundary (bool) - if True, exclude cells on the border of labeled regions

            raw (bool) - if True, use raw measurements from included discs

            use_cache (bool) - if True, used available cached measurement data

        Returns:

            data (pd.Dataframe) - curated cell measurement data, which is None if no measurement data are found

        """

        # load measurements from each stack in the experiment
        data = []
        for stack_id in self.stack_ids:
            stack = self.load_stack(stack_id, full=False)
            measurements = stack.aggregate_measurements(
                selected_only=selected_only,
                exclude_boundary=exclude_boundary,
                raw=raw,
                use_cache=use_cache)

            if measurements is None:
                continue

            # add stack index
            measurements['stack'] = stack._id
            measurements = measurements.set_index('stack', append=True)
            measurements = measurements.reorder_levels([2,0,1])

            data.append(measurements)
            assert stack_id == stack._id, 'Stack IDs do not match.'

        # return None if no data are found
        if len(data) == 0:
            return None

        # aggregate measurements
        data = pd.concat(data, join='outer', sort=False)

        # exclude cells that were not marked for inclusion
        if selected_only:
            data = data[data.selected]

        # exclude cells on clone boundaries
        if exclude_boundary:
            data = data[~data.boundary]

        return data



import numpy as np
import networkx as nx
from copy import deepcopy

from ...visualization import *

from .infomap import InfoMap


class NeighborSampler:
    """
    Class for sampling node attributes averaged over neighbors.

    Attributes:

        graph (spatial.Graph) - graph instance

        G (nx.Graph) - graph with node attribute

        attr (str) - attribute to be averaged over neighbors

        depth (int) - maximum number of edges connecting neighbors

        log (bool) - if True, log-transform values before averaging

    """

    def __init__(self, graph, attr, depth=1, log=True):
        """
        Instantiate sampler for averaging <attr> value over all neighbors within <depth> of each node.

        Args:

            graph (spatial.Graph) - graph instance

            attr (str) - attribute to be averaged over neighbors

            depth (int) - maximum number of edges connecting neighbors

            log (bool) - if True, log-transform values before averaging

        """
        self.graph = graph
        self.attr = attr
        self.depth = depth
        self.log = log

        # add node attribute to graph
        if self.attr_used not in list(self.G.nodes(data=True)[0].keys()):
            self.add_attribute_to_graph()

        # sample over neigbors
        self.average_over_neighbors()

    @classmethod
    def multisample(cls, attr, *graphs, **kwargs):
        """
        Generate composite sample from one or more <graphs>.

        Args:

            attr (str) - attribute to be averaged over neighbors

            graphs (spatial.Graph) - one or more graph instances

            kwargs: keyword arguments for sampler

        Returns:

            sample (np.ndarray[float]) - 2D array of sampled values, first column contains cell measurements while the second column contains measurements averaged over the neighbors of each cell

            keys (list of str) - attribute keys for sampled data

        """
        samples = []
        for graph in graphs:
            sampler = cls(graph, attr, **kwargs)
            samples.append(sampler.sample)
        return np.vstack(samples), sampler.keys

    @property
    def num_nodes(self):
        """ Number of nodes. """
        return self.graph.num_nodes

    @property
    def data(self):
        """ Graph data. """
        return self.graph.data

    @property
    def G(self):
        """ NetworkX graph instance. """
        return self.graph.G

    @property
    def node_values(self):
        """ Vector of attribute values for each node. """
        values = self.data[self.attr].values
        if self.log:
            values = np.log(values)
        return values

    @property
    def node_values_dict(self):
        """ Dictionary of attribute values, keyed by node index. """
        values = self.data[self.attr]
        if self.log:
            values = np.log(values)
        return dict(values)

    @property
    def neighbors(self):
        """ Dictionary of neighbor indices keyed by node indices. """
        kwargs = dict(depth_limit=self.depth)
        bfs = lambda n: [e[1] for e in nx.bfs_edges(self.G, n, **kwargs)]
        return {node: bfs(node) for node in self.G.nodes}

    @property
    def size_attr(self):
        """ Neighborhood size attribute name. """
        return 'd{:d}_neighbors'.format(self.depth)

    @property
    def attr_used(self):
        """ Name of attribute used to access graph data. """
        if self.log:
            return 'log_' + self.attr
        else:
            return self.attr

    @property
    def averaged_attr(self):
        """ Name of averaged attribute. """
        return '{:s}_d{:d}'.format(self.attr, self.depth)

    @property
    def keys(self):
        """ List of attribute names. """
        return [self.attr, self.averaged_attr]

    @property
    def sample(self):
        """ Returns bivariate sample combining each node's attribute value with the average attribute value in its neighborhood. """
        return self.data[self.keys].values

    def add_attribute_to_graph(self):
        """ Add attribute to networkx graph object. """
        nx.set_node_attributes(self.G, self.node_values_dict, name=self.attr_used)

    @staticmethod
    def _neighbor_average(G, node_values, depth=1):
        """
        Average attribute value over all neighbors of each node in a graph.

        Args:

            G (nx.Graph) - graph with nodes containing an attribute

            node_values (dict) - node attribute values keyed by node index

            depth (int) - maximum number of edges connecting neighbors

        Returns:

            node_to_neighborhood (dict) - dictionary of neighborhood (<attribute>, size) tuples keyed by node index

        """

        # define breadth first search of each <src> node
        bfs = lambda n: [e[1] for e in nx.bfs_edges(G, n, depth_limit=depth)]

        # define function to parse mean node attribute within neighborhood
        parse = lambda nbs: (np.mean([node_values[n] for n in nbs]), len(nbs))

        return {src: parse(bfs(src)) for src in G.nodes}

    def average_over_neighbors(self):
        """
        Average attribute value over all neighbors adjacent to each node.
        """

        # average attribute over neighbors
        neighbor_dict = self._neighbor_average(self.G, self.node_values_dict, self.depth)

        # extract average and sample size for each node
        keys, values = list(zip(*neighbor_dict.items()))
        node_indices = np.array(keys)
        means, sizes = np.array(values).T

        # log transform average
        if self.log:
            means = np.exp(means)

        # store outcome
        self.data.loc[node_indices, self.averaged_attr] = means
        self.data.loc[node_indices, self.size_attr] = sizes

    @default_figure
    def histogram_sample_sizes(self, ax=None):
        """ Histogram sample sizes. """
        sizes = self.data[self.size_attr].values
        _ = ax.hist(sizes, bins=np.arange(sizes.max()+1))

    @square_figure
    def plot_neigborhood(self, node,
                         color='r',
                         bg_color='k',
                         ax=None,
                         **kwargs):
        """
        Visualize neighborhood surrounding <node>.

        Args:

            node (int) - node positional index

            color (str or RGB tuple) - color for nodes within neighborhood

            bg_color (str or RGB tuple) - color for nodes outside neighborhood

            kwargs: keyword arguments for plt.scatter

        """
        node_id = self.graph.node_map(node)[0]
        neighbors_id = self.neighbors[node_id]
        neighbors = self.graph.position_map(neighbors_id)

        colors = np.array([bg_color for _ in range(self.graph.nodes.size)])
        colors[node] = color
        colors[neighbors] = color
        ax.scatter(*self.data[self.graph.xykey].values.T, c=colors, **kwargs)

    @default_figure
    def plot_autocorrelation(self, ax=None, xmax=10, **kwargs):
        """ Plot autocorrelation versus path length. """

        # evaluate distance between all nodes
        paths = dict(nx.all_pairs_shortest_path_length(self.G, cutoff=xmax))

        if xmax is None:
            max_depth = max([max(v.values()) for v in paths.values()])
        else:
            max_depth = xmax

        # get node levels and evaluate global mean/variance
        levels = self.node_values
        mu, sigma2 = levels.mean(), levels.var()

        def eval_flux(edge_list):
            """ Evaluate mean fluctuation in edge_list. """
            idx = self.graph.position_map(np.array(edge_list))
            flux = ((levels[idx[:, 0]]-mu) * (levels[idx[:, 1]] - mu)) / sigma2
            return flux.mean(), flux.std(), flux.size

        # compute pairwise fluctuations between nodes
        edges = {i: [] for i in range(max_depth+1)}
        for node, neighbors in paths.items():
            for neighbor, distance in neighbors.items():
                edges[distance].append((node, neighbor))

        # compile autocorrelation function
        means, _, sizes = list(zip(*[eval_flux(e) for e in edges.values()]))
        means, sizes = np.array(means), np.array(sizes)

        # plot autocorrelation
        ax.plot(range(max_depth+1), means, '.-k', **kwargs)
        ax.set_ylim(-0.1, 1)
        ax.set_xlim(0, max_depth+1)
        ax.set_ylabel('Correlation')
        ax.set_xlabel('Path length')


class CommunitySampler(NeighborSampler):
    """
    Class for sampling node attributes averaged over local community.

    Attributes:

        graph (spatial.Graph) - graph instance

        G (nx.Graph) - graph with node attribute

        attr (str) - attribute to be averaged over neighbors

        depth (int) - mean correlation lifetime

        level (int) - hierarchical level at which clusters are merged

        log (bool) - if True, log-transform values before averaging

        twolevel (bool) - if True, use two-level community clustering

    """

    def __init__(self, graph, attr, depth=1., log=True, twolevel=False):
        """
        Instantiate sampler for averaging <attr> value over all neighbors within <depth> of each node.

        Args:

            graph (spatial.Graph) - graph instance

            attr (str) - attribute to be averaged over neighbors

            depth (int) - mean correlation lifetime

            twolevel (bool) - if True, use two-level community clustering

            log (bool) - if True, log-transform values before averaging

        """

        # store attributes
        self.graph = graph
        self.attr = attr
        self.depth = depth
        self.twolevel = twolevel
        self.log = log

        # perform community detection
        if self.graph.imap is None:
            self.graph.detect_communities(twolevel=twolevel)

        # determine clustering level
        self.level = self.clustering_level

        # sample over neigbors
        self.average_over_neighbors()

    @property
    def neighbors(self):
        """ Dictionary of neighbor indices keyed by node indices. """
        labels = self.graph._assign_community(self.level)
        gb = self.data.groupby(labels)
        exclude = lambda node, neighbors: neighbors[neighbors!=node]
        key = lambda n: self.data.index[n]
        return {key(n): exclude(key(n), gb.indices[c]) for n,c in enumerate(labels)}

    @property
    def size_attr(self):
        """ Neighborhood size attribute name. """
        return 'community_size'

    @property
    def averaged_attr(self):
        """ Name of averaged attribute. """
        return '{:s}_community'.format(self.attr)

    @property
    def z_attr(self):
        """ Name of z-scored attribute. """
        return self.attr+'_zscore'

    @property
    def clustering_level(self):
        """ Highest clustering level at which the mean correlation remains above <self.depth> multiples of the decay constant. """
        _, correlations = self.autocorrelate(include_distances=False)
        correlations = np.array(correlations)
        relative_correlation = correlations/correlations[0]
        above = relative_correlation >= np.exp(-self.depth)
        return above.nonzero()[0][-1]

    def average_over_neighbors(self):
        """ Average attribute value over all members of the community encompassing each node. """

        if self.log:
            agg = lambda x: np.log(x).mean()
        else:
            agg = lambda x: x.mean()

        # average over each community
        labels = self.graph._assign_community(self.level)
        community_levels = self.data.groupby(labels)[self.attr].aggregate(agg)
        community_to_mean_level = np.vectorize(dict(community_levels).get)
        means = community_to_mean_level(labels)

        # log transform average
        if self.log:
            means = np.exp(means)

        # evaluate community sizes
        neighbors = self.neighbors
        get_community_size = np.vectorize(lambda n: len(neighbors[n]))

        # store outcome
        self.data[self.averaged_attr] = means
        self.data[self.size_attr] = get_community_size(self.data.index.values)

    def autocorrelate(self, include_distances=False):
        """
        Returns autocorrelation versus community depth.

        Args:

            include_distances (bool) - return mean separate distances

        Returns:

            levels (list) - clustering depths, starting from finest resolution

            correlations (list) - mean correlation within communities

            <optional> distances (list) - mean pairwise separation distance

        """

        # compile z-scored node values
        node_v = self.node_values
        self.data[self.z_attr] = (node_v-node_v.mean())/node_v.std()

        # define hierarchical levels
        levels = list(range(self.graph.imap.aggregator.max_depth))

        # define aggregation functions
        def get_binsize(x):
            """ Evaluate total number of node pairs. """
            return x.size * (x.size - 1) / 2

        def get_fluctuations(x):
            """ Evaluate total pairwise fluctuation between nodes. """
            v = x.values.reshape(-1, 1)
            return self.graph.get_matrix_upper(np.dot(v, v.T)).sum()

        def get_distances(x):
            """ Evaluate total pairwise distance between nodes. """
            distance_matrix = self.graph._distance_matrix(x.values)
            distances = self.graph.get_matrix_upper(distance_matrix)
            return distances.sum()

        # evaluate autocorrelation at each hierarchical level
        correlations, distances = [], []
        for level in levels:
            labels = self.graph._assign_community(level)
            gb = self.data.groupby(labels)
            total_flux = gb[self.z_attr].agg(get_fluctuations).sum()
            num_of_flux = gb[self.z_attr].agg(get_binsize).sum()
            correlations.append(total_flux / num_of_flux)

            if include_distances:
                total_distance = gb[self.graph.xykey].apply(get_distances).sum()
                distances.append(total_distance/num_of_flux)

        if include_distances:
            return distances, correlations
        else:
            return levels, correlations

    @default_figure
    def plot_autocorrelation(self, ax=None, spatial=False):
        """
        Plot autocorrelation versus community depth.

        Args:

            distance (bool) - if True, plot spatial correlation

        """


        levels, correlation = self.autocorrelate(include_distances=spatial)

        # plot autocorrelation
        ax.plot(levels, correlation, '.-k')
        ax.set_ylim(-0.1, 1)
        ax.set_ylabel('Correlation')

        # format axis
        if spatial:
            ax.set_xlabel('Distance')
        else:
            ax.set_xlabel('Hierarchical level')


class RadialSampler(NeighborSampler):
    """
    Class for sampling node attributes averaged within a predetermined radius of each node.

    Attributes:

        graph (spatial.Graph) - graph instance

        G (nx.Graph) - graph with node attribute

        attr (str) - attribute to be averaged over neighbors

        depth (int) - hierarchical level to which communities are merged

        log (bool) - if True, log-transform values before averaging

        length_scale (float) - characteristic length scale of the graph

        radius (float) - radius of sampling region surrounding each measurement

    """

    def __init__(self, graph, attr, depth=1., log=True):
        """
        Instantiate sampler for averaging <attr> value over all nodes within a predetermined radius of each node. The radius is defined by <depth> multiples of the characteristic length over which correlations in the attribute value decay.

        Args:

            graph (spatial.Graph) - graph instance

            attr (str) - attribute to be averaged over neighbors

            depth (int) - hierarchical level to which communities are merged

            log (bool) - if True, log-transform values before averaging

        """

        # store attributes
        self.graph = graph
        self.attr = attr
        self.depth = depth
        self.log = log

        # determine characteristic length scale and set sampling radius
        length_scale = graph.get_correlations(attr, log).characteristic_length

        # if failed (e.g. no correlation), use 1.5x median edge length
        if length_scale is None:
            length_scale = 1.5 * graph.median_edge_length

        self.length_scale = length_scale
        self.radius = depth * length_scale
        self.neighbor_mask = self._neighbor_mask

        # sample over neigbors
        self.average_over_neighbors()

    @property
    def distance_matrix(self):
        """ Euclidean distance matrix between nodes (ordered by position in <self.data>). """
        return self.graph.distance_matrix

    @property
    def _neighbor_mask(self):
        """ Boolean adjacency mask (True for neighbors). """
        neighbor_mask = (self.distance_matrix < self.radius)
        np.fill_diagonal(neighbor_mask, False)
        return neighbor_mask

    @property
    def neighbors(self):
        """
        Dictionary of neighbor positional indices keyed by node indices.
        """
        return  {n: r.nonzero()[0] for n, r in enumerate(self.neighbor_mask)}

    @property
    def size_attr(self):
        """ Neighborhood size attribute name. """
        return 'sampling_radius'

    @property
    def averaged_attr(self):
        """ Name of averaged attribute. """
        return '{:s}_radial'.format(self.attr)

    def average_over_neighbors(self):
        """
        Average attribute value over all nodes within the specified radius of each node.
        """

        # average within each neighborhood
        node_values = self.node_values.reshape(1, -1)
        matrix = np.repeat(node_values, self.num_nodes, axis=0)
        masked_values = np.ma.masked_array(matrix, mask=~self.neighbor_mask)
        means = masked_values.mean(axis=1).data

        # log transform average
        if self.log:
            means = np.exp(means)

        # store outcome
        self.data[self.averaged_attr] = means
        self.data[self.size_attr] = (~masked_values.mask).sum(axis=1)

    @square_figure
    def plot_neigborhood(self, node,
                         color='r',
                         bg_color='k',
                         radius=True,
                         ax=None,
                         **kwargs):
        """
        Visualize neighborhood surrounding <node>.

        Args:

            node (int) - node positional index

            color (str or RGB tuple) - color for nodes within neighborhood

            bg_color (str or RGB tuple) - color for nodes outside neighborhood

            radius (bool) - if True, plot sampling radius

            kwargs: keyword arguments for plt.scatter

        """

        # draw selection boundary
        if radius:
            center = self.data[self.graph.xykey].values[node]
            circle = plt.Circle(center, self.radius, color=color, alpha=0.2)
            ax.add_artist(circle)

        # scatter points
        colors = np.array([bg_color for _ in range(self.graph.nodes.size)])
        colors[node] = color
        colors[self.neighbors[node]] = color
        ax.scatter(*self.data[self.graph.xykey].values.T, c=colors, **kwargs)

    @default_figure
    def plot_autocorrelation(self, ax=None, **kwargs):
        """ Plot autocorrelation versus community depth. """
        correlations = self.graph.get_correlations(self.attr, self.log)
        correlations.visualize(ax=ax, **kwargs)



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
from scipy.spatial import Voronoi
from matplotlib.collections import PolyCollection
from matplotlib.colors import Normalize

from .spatial.alpha import AlphaShapes


class CloneBoundaries:
    """
    Object for drawing paths around clone boundaries.

    Attributes:

        shapes (list of AlphaShapes) - shape indices for ordered labels

    """

    def __init__(self, graph, label_by='genotype', alpha=50):
        """
        Instantiate clone boundary object.

        Args:

            graph (spatial.Graph)

            label_by (str) - attribute used to label clones

            alpha (float) - shape parameter for alpha shapes

        """
        self.shapes = self.build_shapes(graph, label_by, alpha)

    @classmethod
    def from_layer(cls, layer, label_by='genotype', **kwargs):
        """ Instantiate from clones.Layer instance. """
        return cls(layer.graph, label_by=label_by, **kwargs)

    @property
    def norm(self):
        """ Normalization for number of shapes. """
        return Normalize(vmin=0, vmax=len(self.shapes)-1)

    @staticmethod
    def build_shapes(graph, label_by, alpha):
        """ Compile shapes. """
        shapes_dict = {}
        for label in graph.data[label_by].unique():
            xy = graph.data[graph.data[label_by]==label][graph.xykey].values
            shapes_dict[label] = AlphaShapes(xy, alpha=alpha)
        return shapes_dict

    def plot_boundary(self, label, **kwargs):
        """ Plot boundary for clones with <label>. """
        shape = self.shapes[label]
        shape.plot_boundary(**kwargs)

    def plot_boundaries(self, cmap=plt.cm.viridis, **kwargs):
        """ Plot all clone boundaries. """
        for label in self.shapes.keys():
            color = cmap(self.norm(label))
            self.plot_boundary(label, color=color, **kwargs)


class Tessellation:
    """
    Object for visualizing Voronoi tessellations.
    """

    def __init__(self, xy, labels, q=90, colors=None):

        self.vor = Voronoi(xy)
        self.vor.regions = np.array(self.vor.regions)
        self.set_region_mask(q=q)
        self.region_labels = self.label_regions(labels)
        self.verts = self.vor.regions[self.mask]
        self.set_cmap(colors)

    def label_regions(self, labels):
        points = np.argsort(self.vor.point_region)
        point_to_label = np.vectorize(dict(enumerate(labels)).get)
        region_labels = point_to_label(points)
        return region_labels

    def set_cmap(self, colors=None):
        N = len(set(self.region_labels))
        if colors is None:
            colors = np.random.random((N, 3))
        self.cmap = ListedColormap(colors, 'indexed', N)

    @staticmethod
    def _evaluate_area(x, y):
        """ Evaluate area enclosed by a set of points. """
        return 0.5*np.abs(np.dot(x, np.roll(y,1))-np.dot(y, np.roll(x,1)))

    def evaluate_region_area(self, region):
        """ Evaluate pixel area enclosed by a region. """
        return self._evaluate_area(*self.vor.vertices[region, :].T)

    def set_region_mask(self, q=90):
        """
        Mask regions with pixel areas larger than a specified quantile.

        Args:
        q (float) - maximum region area quantile, 0 to 100
        """
        f = np.vectorize(lambda x: -1 not in x and len(x) > 0)
        mask = f(self.vor.regions)
        mask *= self.build_region_area_mask(q=q)
        self.mask = mask

    def build_region_area_mask(self, q=90):
        """
        Mask regions with pixel areas larger than a specified quantile.

        Args:

            q (float) - maximum region area quantile, 0 to 100

        Returns:

            mask (np.ndarray[bool]) - True for regions smaller than maximum area

        """
        evaluate_area = np.vectorize(lambda x: self.evaluate_region_area(x))
        areas = evaluate_area(self.vor.regions)
        threshold = np.percentile(areas, q=q)
        return (areas <= threshold)

    @staticmethod
    def _show(vertices, c='k', ax=None, alpha=0.5):
        """ Visualize vertices. """
        if ax is None:
            fig, ax = plt.subplots()
            ax.set_xlim(0, 2048)
            ax.set_ylim(0, 2048)
            ax.axis('off')
        poly = PolyCollection(vertices)
        poly.set_facecolors(c)
        poly.set_alpha(alpha)
        ax.add_collection(poly)

    def show(self, ax=None, **kw):
        """ Visualize vertices. """
        get_vertices = np.vectorize(lambda region: self.vor.vertices[region])
        vertices = [self.vor.vertices[r] for r in self.vor.regions[self.mask]]
        c = self.cmap(self.region_labels[self.mask[1:]])
        self._show(vertices, c=c, ax=ax, **kw)


class CloneVisualization(Tessellation):
    """
    Object for visualizing clones by shading Voronoi cells.
    """

    def __init__(self, graph, label='genotype', **kw):
        labels = graph.data[label].values
        Tessellation.__init__(self, graph.node_positions_arr, labels, **kw)



from os.path import join
import numpy as np
import matplotlib.pyplot as plt

from ..visualization import *


class BenchmarkingResults:
    """
    Container for managing aggregated results of a benchmarking sweep.
    """

    def __init__(self, data, shape,
                 methods=None,
                 clone_sizes=None,
                 ambiguities=None):
        """
        Instantiate benchmarking results object.

        Args:

            data (pd.DataFrame) - benchmarking performance data

            shape (tuple) - shape of benchmarking sweep

            methods (array like) - methods included

            clone_sizes (array like) - mean size of clones (x-values)

            ambiguities (array like) - fluorescence ambiguities (y-values)

        """

        if methods is None:
            methods = ['labels_MAE',
                'level_only_MAE',
                'spatial_only_MAE',
                'community_MAE',
                'labels_PCT',
                'level_only_PCT',
                'spatial_only_PCT',
                'community_PCT']
            methods = [m for m in methods if m in data.columns]

        self.methods = {m: i for i, m in enumerate(methods)}

        columns = ['row_id', 'column_id', 'ambiguity_id']
        self.data = data.groupby(columns)[methods].mean()
        self.shape = shape

        self.clone_sizes = clone_sizes
        self.ambiguities = ambiguities

    @property
    def num_methods(self):
        """ Number of methods. """
        return len(self.methods)

    def slice(self, row_id=0):
        """
        Returns score matrices for the specified <row_id>. Row IDs correspond to recombination rates.
        """
        data = self.data.xs(row_id, level=0).swaplevel().sort_index(level=0)
        shape = [len(s) for s in data.index.levels]
        matrices = data.values.reshape(*shape, self.num_methods)
        return matrices

    @staticmethod
    def build_figure(nrows=1, ncols=1, figsize=(2, 2)):
        """ Returns figure and axis. """
        return plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize)

    @staticmethod
    def plot_image(ax, im, fliplr=False, flipud=False, **kwargs):
        """ Plot <im> on <ax>. """

        if fliplr:
            im = np.fliplr(im)

        if flipud:
            im = np.flipud(im)

        ax.imshow(im, **kwargs)

    @default_figure
    def plot_relative_error(self,
                            method='community_MAE',
                            reference_method='level_only_MAE',
                            row_id=0,
                            vmin=-3,
                            vmax=3,
                            cmap=plt.cm.seismic,
                            ax=None,
                            figsize=(2., 2.),
                            **kwargs):
        """
        Plots relative error rate for a given <row_id>.

        """

        scores = self.slice(row_id)[:,:, self.methods[method]]
        reference_scores = self.slice(row_id)[:,:, self.methods[reference_method]]
        foldchange = np.log2(scores/reference_scores)

        kw = dict(vmin=vmin, vmax=vmax, cmap=cmap)
        kw.update(kwargs)
        self.plot_image(ax, foldchange, fliplr=True, flipud=False, **kw)
        self.format_ax(ax)

    @default_figure
    def plot_absolute_error(self,
                            method='community_MAE',
                            row_id=0,
                            log=True,
                            vmin=0.,
                            vmax=0.33,
                            cmap=plt.cm.inferno,
                            ax=None,
                            figsize=(2., 2.),
                            **kwargs):
        """
        Plots absolute error rates for a given <row_id> and <method>.
        """

        matrix = self.slice(row_id)[:,:,self.methods[method]]

        # log-transform values
        if log:
            matrix = np.log10(matrix)
            vmin, vmax = np.log10(vmin), np.log10(vmax)

        # plot
        kw = dict(vmin=vmin, vmax=vmax, cmap=cmap)
        kw.update(kwargs)
        self.plot_image(ax, matrix, fliplr=True, flipud=False, **kw)
        self.format_ax(ax)

    def format_ax(self, ax):
        """ Apply formatting to axis. """

        if self.clone_sizes is not None:
            xvals = self.clone_sizes[::-1]
        else:
            xvals = ax.get_xticks()

        if self.ambiguities is not None:
            yvals = self.ambiguities
        else:
            yvals = ax.get_yticks()

        # set axis extent
        extent = [xvals.min(), xvals.max(), yvals.max(), yvals.min()]
        ax.images[0].set_extent(extent)
        ax.invert_yaxis()
        ax.set_aspect(xvals.ptp()/yvals.ptp())

        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.set_ylabel('Fluorescence ambiguity')
        ax.set_xlabel('Clone size')



from os.path import join
import matplotlib.pyplot as plt

from .interface import StackInterface


class GUI(StackInterface):
    """
    GUI for manually selecting regions within each layer of an image stack.

    Following instantiation, run "GUI.connect()" to begin event handling.

    Key actions:

        T: remove last added point

        Y: reset all points in layer

        E: exclude entire layer

        W: save ROI data

        Q: disconnect and exit GUI

    """

    def __init__(self, stack):
        """
        Instantiate GUI.

        Args:

            stack (Stack) - image stack

        """
        super().__init__(stack)

        # set attributes
        self.traceback = []
        self.saved = False

    @staticmethod
    def load(stack):
        """
        Load selection GUI from file.

        Args:

            stack (Stack) - image stack

        Returns:

            gui (GUI) - disconnected gui

        """
        gui = GUI(stack)
        _ = [interface.load() for interface in gui.layer_to_interface.values()]
        return gui

    def save(self, image=True):
        """
        Save selection path for each layer.

        Args:

            image (bool) - if True, save overall image of selections

        """

        # save each layer
        for layer_gui in self.layer_to_interface.values():
            layer_gui.clear_markers()
            layer_gui.save()

        # save selection image
        if image:
            kw = dict(format='png', dpi=200, bbox_inches='tight', pad_inches=0)
            im_path = join(self.path, 'selection.png')
            self.fig.savefig(im_path, **kw)

        self.saved = True

    def connect(self):
        """ Connect event handling. """
        self.cid_click = self.fig.canvas.mpl_connect('button_press_event', self.on_click)
        self.cid_press = self.fig.canvas.mpl_connect('key_press_event', self.on_key)

    def disconnect(self):
        """ Disconnect event handling. """
        self.fig.canvas.mpl_disconnect(self.cid_click)
        self.fig.canvas.mpl_disconnect(self.cid_press)

    def exit(self):
        """ Disconnect and exit GUI. """
        self.disconnect()
        self.fig.clf()
        plt.close(self.fig)

    def which_layer(self, event):
        """ Returns layer ID where event took place. """
        return self.layer_to_interface[self.ax_to_layer[event.inaxes]]

    def on_click(self, event):
        """ Click action: add point. """
        row = self.which_layer(event)
        pt = (event.xdata, event.ydata)
        if None not in pt:
            row.add_point(pt)

    def on_key(self, event):
        """
        Key actions.

            T: remove last added point

            Y: reset all points in layer

            E: exclude entire layer

            W: save ROI data

            Q: disconnect and exit GUI

        """

        row = self.which_layer(event)

        # save and disconnect
        if event.key == 'w':
            self.save()

        # save and disconnect
        elif event.key == 'q':
            try:
                self.exit()
            except Exception as error:
                self.traceback.append(error)

        # mark for exclusion
        elif event.key == 'e':
            row.include = False
            row.overlay('EXCLUDED')

        # undo
        elif event.key == 't':
            row.undo()

        # clear
        elif event.key == 'y':
            row.clear()



import warnings
from os.path import join, exists, abspath, isdir
from os import mkdir, remove
from shutil import move, rmtree
from glob import glob
import gc
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from ..utilities import IO
from ..utilities import UserPrompts
from ..annotation import Annotation

from .layers import Layer
from .silhouette_write import WriteSilhouette

# filter numpy warnings
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")


class StackIO(WriteSilhouette):
    """ Methods for saving and loading a Stack instance. """

    @staticmethod
    def from_tif(filepath, bit_depth):
        """
        Initialize stack from tif <filepath>.

        Args:

            path (str) - path to tif image file

            bit_depth (int) - bit depth of raw tif (e.g. 12 or 16)

        Returns:

            stack (flyqma.Stack)

        """

        path, ext = filepath.rsplit('.', maxsplit=1)
        if ext.lower() != 'tif':
            raise ValueError('TIF extension not found.')
        _id = path.split('/')[-1]

        # make directory
        mkdir(path)
        move(filepath, join( path, '{:s}.tif'.format(_id)))

        return Stack(path, bit_depth=bit_depth)

    @staticmethod
    def from_silhouette(filepath, bit_depth):
        """
        Initialize stack from silhouette <filepath>.

        Args:

            path (str) - path to silhouette file

            bit_depth (int) - bit depth of raw tif (e.g. 12 or 16)

        Returns:

            stack (flyqma.Stack)

        """

        raise UserWarning('INCOMPLETE METHOD: TIF FILE REQUIRED.')

        path, ext = filepath.rsplit('.', maxsplit=1)
        assert ext.lower() == 'silhouette', 'Silhouette file not found.'
        _id = path.split('/')[-1]

        # make directory
        mkdir(path)
        move(filepath, join(path, '{:s}.silhouette'.format(_id)))

        return Stack(path, bit_depth=bit_depth)

    def save(self):
        """ Save stack metadata and annotator. """
        self.save_metadata()
        if self.annotator is not None:
            self.save_annotator()

    def save_metadata(self):
        """ Save metadata. """
        io = IO()
        io.write_json(join(self.path, 'metadata.json'), self.metadata)

    def save_annotator(self, data=True):
        """ Save annotator to annotation directory. """
        if not isdir(self.annotator_path):
            mkdir(self.annotator_path)
        self.annotator.save(self.annotator_path, data=data)

    def load_metadata(self):
        """ Load available metadata. """
        metadata_path = join(self.path, 'metadata.json')
        if exists(metadata_path):
            io = IO()
            self.metadata = io.read_json(metadata_path)

    def load_annotator(self):
        """ Load annotator from annotation directory. """
        if exists(self.annotator_path):
            self.annotator = Annotation.load(self.annotator_path)

    def load_image(self):
        """ Load 3D image from tif file. """

        # load tif, normalize pixel intensities, and convert to NWHC format
        stack = self._read_tif(self.tif_path)
        stack = self._pixel_norm(stack, self.bit_depth)
        stack = self._to_NWHC(stack)
        self.stack = stack

        # set stack shape
        self.metadata['depth'] = self.stack.shape[0]
        self.metadata['colordepth'] = self.stack.shape[-1]

    @staticmethod
    def _read_tif(path):
        """ Read tif from <path>. """
        return IO().read_tiff(path)

    @staticmethod
    def _pixel_norm(pixels, bitdepth=12):
        """ Normalize <pixels> intensities by <bitdepth>. """
        return pixels / (2**bitdepth)

    @staticmethod
    def _reorder_channels(stack, idx):
        """ Reorder channels in <stack> according to <idx>. """
        return stack[:, :, :, idx]

    @staticmethod
    def _to_NWHC(stack):
        """
        Convert image to NWHC format. Native format is automatically detected using some heuristics regarding image size relative to channel depth.

        Args:

            stack (np.ndarray[float]) - original image stack

        Returns:

            stack_NWHC (np.ndarray[float]) - image stack in NWHC format

        """

        # if only one layer is provided, append depth dimension (N)
        if len(stack.shape) == 3:
            stack = stack.reshape(1, *stack.shape)

        # determine channel dimension (C) - assumes it's smaller than W & H
        c_dim = np.argmin(stack.shape[1:]) + 1

        # swap axes until C is the last dimension
        while c_dim != 3:
            stack = np.swapaxes(stack, c_dim, c_dim+1)
            c_dim += 1

        return stack


class Stack(StackIO):
    """
    Object represents a 3D RGB image stack.

    Attributes:

        path (str) - path to stack directory

        _id (str) - stack ID

        stack (np.ndarray[float]) - 3D RGB image stack

        shape (tuple) - stack dimensions, (depth, X, Y, 3)

        bit_depth (int) - bit depth of raw tif image

        stack_depth (int) - number of layers in stack

        color_depth (int) - number of fluorescence channels in stack

        annotator (Annotation) - object that assigns labels to measurements

        metadata (dict) - stack metadata

        tif_path (str) - path to multilayer RGB tiff file

        layers_path (str) - path to layers directory

        annotator_path (str) - path to annotation directory

    """

    def __init__(self, path, bit_depth=None):
        """
        Initialize stack from stack directory <path>.

        Args:

            path (str) - path to stack directory

            bit_depth (int) - bit depth of raw tif (e.g. 12 or 16). Value will be read from the stack metadata if None is provided. An error is raised if no value is found.

        """

        # strip trailing slashes
        path = path.rstrip('/')

        # check if path is directly to a silhouette file
        if '.silhouette' in path.lower():
            raise ValueError('Please use the Stack.from_silhouette constructor.')
        elif '.tif' in path.lower():
            raise ValueError('Please use the Stack.from_tif constructor.')

        # set path to stack directory
        self._id = path.rsplit('/', maxsplit=1)[-1]
        self.path = abspath(path)
        self.stack = None

        # find image file (defaults to stack name, otherwise first tif found)
        tifs = list(glob(join(path, '*.tif'), recursive=False))
        assert len(tifs) > 0, 'No tif image files found in stack directory.'
        default_path = join(path, '{:s}.tif'.format(self._id))
        if default_path in tifs:
            self.tif_path = default_path
        else:
            self.tif_path = tifs[0]

        # set layer and annotator directory paths
        self.layers_path = join(self.path, 'layers')
        self.annotator_path = join(self.path, 'annotation')

        # initialize stack
        if not self.is_initialized:
            if type(bit_depth) == int:
                self.initialize(bit_depth=bit_depth)
            else:
                self.prompt_initialization()

        # load metadata
        self.load_metadata()

        # if bit_depth was provided, make sure it's consistent
        if bit_depth is None:
            msg = 'bit_depth was neither specified nor found in the metadata.'
            assert 'bits' in self.metadata.keys(), msg
        else:
            msg = 'Specified bit depth is inconsistent with existing metadata.'
            assert self.bit_depth == bit_depth, msg

        # load annotator
        self.annotator = None
        self.load_annotator()

        # reset layer iterator count
        self.count = 0

    def __getitem__(self, layer_id):
        """ Load layer. """
        return self.load_layer(layer_id, graph=True, use_cache=True, full=True)

    def __iter__(self):
        """ Iterate across included layers. """
        self.count = 0
        return self

    def __next__(self):
        """ Return next included layer. """
        if self.count < len(self.included):
            layer_id = self.included[self.count]
            layer = self.__getitem__(layer_id)
            self.count += 1
            return layer
        else:
            raise StopIteration

    def prompt_initialization(self):
        """ Ask user whether to initialize all stack directories. """
        msg = '{:s} directory has not been initialized. Do it now?'.format(self.filename)
        user_response = UserPrompts.boolean_prompt(msg)
        if user_response:
            msg = 'Please enter an image bit depth:'
            bit_depth = UserPrompts.integer_prompt(msg)
            if bit_depth is not None:
                self.initialize(bit_depth=bit_depth)
            else:
                raise ValueError('Could not initialize stack because bit depth value was not recognized.')

    @staticmethod
    def _check_if_initialized(path):
        """ Returns True if <path> contains an initialized stack directory. """
        layers_complete = isdir(join(path, 'layers'))
        metadata_complete = exists(join(path, 'metadata.json'))
        if not (layers_complete and metadata_complete):
            return False
        else:
            return True

    @property
    def is_initialized(self):
        """ Returns True if Stack has been initialized. """
        return self._check_if_initialized(self.path)

    @property
    def is_segmented(self):
        """ True if segmentation is complete. """
        return self.aggregate_measurements() is not None

    @property
    def is_annotated(self):
        """ True if annotation is complete. """
        return self.annotator is not None

    @property
    def included(self):
        """ Indices of included layers. """
        return self.get_included_layers()

    @property
    def filename(self):
        """ Stack filename. """
        return self.path.split('/')[-1]

    @property
    def bit_depth(self):
        """ Bit depth of raw image. """
        return self.metadata['bits']

    @property
    def stack_depth(self):
        """ Number of layers in stack. """
        return self.metadata['depth']

    @property
    def color_depth(self):
        """ Number of fluorescence channels in stack. """
        return self.metadata['colordepth']

    @property
    def selector_path(self):
        """ Path to model selection object. """
        return join(self.annotator_path, 'models')

    def get_included_layers(self):
        """ Returns indices of included layers. """
        layers = [self.load_layer(i, graph=False) for i in range(self.stack_depth)]
        return [layer._id for layer in layers if layer.include]

    def initialize(self, bit_depth):
        """
        Initialize stack directory.

        Args:

            bit_depth (int) - bit depth of raw tif (e.g. 12 or 16)

        """

        # make layers directory
        if not exists(self.layers_path):
            mkdir(self.layers_path)

        # load stack (to determine shape), and check bit depth
        raw = self._read_tif(self.tif_path)
        assert raw.max() <= (2**bit_depth), 'Pixels exceed bit_depth.'
        stack = raw / (2**bit_depth)

        # make metadata file
        io = IO()
        self.metadata = dict(bits=bit_depth, depth=stack.shape[0], params={})
        io.write_json(join(self.path, 'metadata.json'), self.metadata)

        # load image
        if self.stack is None:
            self.load_image()

        # initialize layers
        for layer_id in range(self.stack_depth):
            layer_path = join(self.layers_path, '{:d}'.format(layer_id))
            layer = Layer(layer_path)
            layer.initialize()

    def restore_directory(self):
        """ Restore stack directory to original state. """

        dirs = [
            self.layers_path,
            join(self.path, 'annotation'),
            join(self.path, 'metadata.json')]

        for path in dirs:
            if exists(path):
                if isdir(path):
                    rmtree(path)
                else:
                    remove(path)

    def segment(self, channel,
                preprocessing_kws={},
                seed_kws={},
                seg_kws={},
                min_area=250,
                save=True):
        """
        Segment all layers using watershed strategy.

        Args:

            channel (int) - channel index on which to segment image

            preprocessing_kws (dict) - keyword arguments for image preprocessing

            seed_kws (dict) - keyword arguments for seed detection

            seg_kws (dict) - keyword arguments for segmentation

            min_area (int) - threshold for minimum segment size, px

            save (bool) - if True, save measurement data for each layer

        """

        # make sure image is loaded
        assert self.stack is not None, 'Image data not found. Use the load_image() method to read image data into memory.'

        for layer in self:
            _ = layer.segment(channel,
                preprocessing_kws=preprocessing_kws,
                seed_kws=seed_kws,
                seg_kws=seg_kws,
                min_area=min_area)

            # save layer measurements and segmentation
            if save:
                layer.save()

    def train_annotator(self, attribute,
                        save=False,
                        logratio=True,
                        num_labels=3,
                        **kwargs):
        """
        Train an Annotation model on all layers in this stack.

        Args:

            attribute (str) - measured attribute used to determine labels

            save (bool) - if True, save annotator and model selection routine

            logratio (bool) - if True, weight edges by relative attribute value

            num_labels (int) - number of allowable unique labels

            kwargs: keyword arguments for Annotation, including:

                sampler_type (str) - either 'radial', 'neighbors', 'community'

                sampler_kwargs (dict) - keyword arguments for sampler

                min_num_components (int) - minimum number of mixture components

                max_num_components (int) - maximum number of mixture components

                addtl_kwargs: keyword arguments for Classifier

        """

        # make sure measurement data are available
        if not self.is_segmented:
            raise RuntimeError('Measurement data not available, please run segmentation first.')

        # build graph for each layer
        graphs = []
        for layer_id in self.included:
            layer = self.load_layer(layer_id, False, False, False)
            layer.build_graph(attribute, logratio=logratio)
            graphs.append(layer.graph)

            # save graph metadata
            if save:
                layer.save_metadata()

        # train annotator
        annotator = Annotation(attribute, num_labels=num_labels, **kwargs)
        selector = annotator.train(*graphs)
        self.annotator = annotator

        # save models
        if save:
            self.save_annotator(data=True)
            selector.save(self.annotator_path)

            # annotate measurements
            for layer in self:
                layer.annotate()
                layer.save_processed_data()

    def aggregate_measurements(self,
                               selected_only=False,
                               exclude_boundary=False,
                               raw=False,
                               use_cache=True):
        """
        Aggregate measurements from each included layer.

        Args:

            selected_only (bool) - if True, exclude cells not marked for inclusion

            exclude_boundary (bool) - if True, exclude cells on clone boundaries

            raw (bool) - if True, aggregate raw measurements

            use_cache (bool) - if True, used available cached measurement data

        Returns:

            data (pd.Dataframe) - measurement data (None if unavailable)

        """

        # load measurements from each included layer
        data = []
        for layer_id in self.included:
            layer = self.load_layer(layer_id,
                                    graph=(not use_cache),
                                    use_cache=use_cache,
                                    full=False)

            # skip layers without measurements
            if not layer.is_segmented:
                continue

            # get raw or processed measurements
            if raw:
                layer_data = layer.measurements
            else:
                layer_data = layer.data

            layer_data['layer'] = layer._id
            data.append(layer_data)
            assert layer_id == layer._id, 'Layer IDs do not match.'

        # return None if no data are found
        if len(data) == 0:
            return None

        # aggregate measurement data
        data = pd.concat(data, join='outer', sort=False)
        data = data.set_index(['layer', 'segment_id'])

        # exclude cells outside the ROI
        if selected_only:
            assert 'selected' in data.columns, 'ROI not defined.'
            data = data[data.selected]

        # exclude cells on the border of labeled regions
        if exclude_boundary:
            assert 'boundary' in data.columns, 'Cannot exclude boundary regions because no regions have been defined. Annotate the stack then try again.'
            data = data[~data.boundary]

        # load manual labels from silhouette
        if exists(self.silhouette_path):
            data = data.join(self.load_silhouette_labels())

        return data

    def load_layer(self, layer_id=0, graph=True, use_cache=True, full=True):
        """
        Load individual layer.

        Args:

            layer_id (int) - layer index

            graph (bool) - if True, load layer graph

            use_cache (bool) - if True, use cached layer measurement data

            full (bool) - if True, load fully labeled RGB image

        Returns:

            layer (Layer)

        """

        # define layer path
        layer_path = join(self.layers_path, '{:d}'.format(layer_id))

        # if performing light load, don't pass image
        if full and self.stack is not None:
            im = self.stack[layer_id, :, :, :]
        else:
            im = None

        # instantiate layer
        layer = Layer(layer_path, im, self.annotator)

        # load layer
        layer.load(use_cache=use_cache, graph=graph)

        return layer



from os.path import join, exists
from time import time
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from ..utilities import Pickler
from ..annotation import Annotation

from .training import Training
from .simulation import SimulationBenchmark


class BatchBenchmark(Pickler, Training):
    """
    Class for benchmarking a batch of simulations.

    Attributes:

        batch (growth.sweep.batch.Batch) - batch of simulations

        ambiguity (float) - fluorescence ambiguity

        num_replicates (int) - number of fluorescence replicates

        data (pd.DataFrame) - synthetic measurement data

        attribute (str) - attribute on which cell measurements are classified

        graphs (dict) - graph objects for each replicate, keyed by replicate_id

        annotator (Annotation) - object that assigns labels to measurements

        results (pd.DataFrame) - classification scores for each replicate

        runtime (float) - total benchmark evaluation runtime

    """

    def __init__(self, batch,
                 ambiguity=0.1,
                 num_replicates=1,
                 attribute='clonal_marker',
                 logratio=True,
                 train_globally=True,
                 training_kw={},
                 testing_kw={}):
        """
        Instantiate batch benchmark.

        Args:

            batch (growth.sweep.batch.Batch) - batch of simulations

            ambiguity (float) - fluorescence ambiguity

            num_replicates (int) - number of fluorescence replicates

            attribute (str) - attribute on which measurements are classified

            logratio (bool) - if True, weight graph edges by log-ratio of attribute level. otherwise, use the absolute difference

            train_globally (bool) - if True, train annotator on entire batch

            training_kw (dict) - keyword arguments for annotator training

            testing_kw (dict) - keyword arguments for annotator application

        """
        self.batch = batch
        self.ambiguity = ambiguity
        self.num_replicates = num_replicates
        self.attribute = attribute
        self.logratio = logratio
        self.train_globally = train_globally

        self.training_kw = training_kw
        self.testing_kw = testing_kw

        # initialize attributes
        self.data = None
        self.annotator = None
        self.results = None
        self.runtime = None

    def __getitem__(self, replicate_id):
        """ Returns SimulationBenchmark for <replicate_id>. """
        measurements = self.data.iloc[self.replicates.indices[replicate_id], :]
        return SimulationBenchmark(measurements.copy(), **self.params)

    def save(self, dirpath, data=False):
        """
        Save serialized batch instance to <dirpath>.

        Args:

            dirpath (str) - destination of serialized object

            data (bool) - if True, include measurement data (training data)

        """
        if not data:
            self.data = None
            self.graphs = None

        if self.annotator is not None:
            self.annotator.save(dirpath, data=data)
            self.annotator = None

        super().save(join(dirpath, 'batch.pkl'))

    @staticmethod
    def load(dirpath):
        """ Load batch from <dirpath>. """

        batch = Pickler.load(join(dirpath, 'batch.pkl'))

        # load annotator
        if exists(join(dirpath, 'annotation.json')):
            annotator = Annotation.load(dirpath)
            batch.annotator = annotator

        return batch

    @property
    def params(self):
        """ Parameters for SimulationBenchmark. """
        return dict(annotator=self.annotator,
                    attribute=self.attribute,
                    logratio=self.logratio,
                    training_kw=self.training_kw,
                    testing_kw=self.testing_kw)

    @property
    def multiindex(self):
        """ Multilevel index for replicates. """
        return ['growth_replicate', 'fluorescence_replicate']

    @property
    def replicates(self):
        """ Replicates iterator (pd.GroupBy) """
        return self.data.groupby(self.multiindex)

    def build_graphs(self):
        """ Build graph objects for each replicate. """
        kw = dict(weighted_by=self.attribute, logratio=self.logratio)
        return {_id: self.build_graph(r, **kw) for _id, r in self.replicates}

    def measure(self):
        """ Returns dataframe of synthetic measurements. """
        return self.batch.measure(self.ambiguity, self.num_replicates)

    def evaluate_benchmarks(self):
        """
        Evaluate benchmark for each replicate.
        """

        # iterate over replicates
        results = {}
        for replicate_id, replicate in self.replicates:

            # evaluate benchmark for current replicate
            bmark = SimulationBenchmark(replicate.copy(),
                                            graph=self.graphs[replicate_id],
                                            **self.params)

            # store results
            results[replicate_id] = dict(

                labels_MAE=bmark.scores['labels'].MAE,
                level_only_MAE=bmark.scores['level_only'].MAE,
                spatial_only_MAE=bmark.scores['spatial_only'].MAE,
                community_MAE=bmark.scores['labels_comm'].MAE,

                labels_PCT=bmark.scores['labels'].percent_correct,
                level_only_PCT=bmark.scores['level_only'].percent_correct,
                spatial_only_PCT=bmark.scores['spatial_only'].percent_correct,
                community_PCT=bmark.scores['labels_comm'].percent_correct)

        # compile dataframe
        results = pd.DataFrame.from_dict(results, orient='index')
        results.index.set_names(self.multiindex, inplace=True)

        return results

    def run(self, train=True):
        """
        Run benchmark on batch.

        Args:

            train (bool) - if True, train global classifier

        """

        # generate synthetic measurements
        start = time()
        self.data = self.measure()

        # build graphs
        self.graphs = self.build_graphs()

        # train annotation object
        if train and self.train_globally:
            self.annotator = self.train(*list(self.graphs.values()),
                                        attribute=self.attribute,
                                        **self.training_kw)

        elif not self.train_globally:
            self.annotator = None

        # evaluate benchmarks
        self.results = self.evaluate_benchmarks()
        self.runtime = time() - start

    def benchmark_simulation(self, replicate_id=0):
        """
        Returns SimulationBenchmark for <replicate_id>.

        This is separate from the large batch of measurement data generated by the run method because the annotator is only trained on an individual replicate.

        """
        sim = self.batch[replicate_id]
        measurements = sim.measure(ambiguity=self.ambiguity)
        return SimulationBenchmark(measurements, **self.params)



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from scipy.stats import norm, multivariate_normal

from .univariate import UnivariateMixture
from .visualization import BivariateVisualization


class BivariateMixtureProperties:
    """ Extension properties for bivariate mixtures. """

    @property
    def supportx(self):
        """ Support for first dimension. """
        return super().support
        #max_val = np.percentile(self.values[:, 0], q=99.9)
        #return np.linspace(self.values[:, 0].min(), max_val, num=100)

    @property
    def supporty(self):
        return self.supportx

    @property
    def support(self):
        xx, yy = np.meshgrid(self.supportx, self.supporty)
        return np.vstack((xx.ravel(), yy.ravel())).T

    @property
    def extent(self):
        """ Extent for x and y axes. """
        return np.array([self.lbound, self.ubound, self.ubound, self.lbound])

    @property
    def scale_factor(self):
        return np.exp(self.support).reshape(self.support_size[0], self.support_size[1], self.dim)

    @property
    def support_size(self):
        return (self.supportx.size, self.supporty.size)

    @property
    def components(self):
        return [multivariate_normal(mean=self.means_[i], cov=self.covariances_[i]) for i in range(self.n_components)]


class BivariateMixture(BivariateMixtureProperties,
                       BivariateVisualization,
                       UnivariateMixture):
    """
    Bivariate Gaussian mixture model.

    Inherited attributes:

        values (array like) - values to which model was fit

        See sklearn.mixture.GaussianMixture

    """

    dim = 2

    def __getitem__(self, margin):
        """ Returns univariate mixture model for specified <margin>. """
        return self.get_marginal_mixture(margin)

    def get_marginal_mixture(self, margin):
        """ Returns univariate mixture model for specified <margin>. """

        values = self.values[:, [margin]]
        mu = self.means_[:, [margin]]
        cov = self.covariances_[:, [margin]]
        weights = self.weights_

        args = (mu, cov, weights)
        kwargs = dict(values=values)

        return UnivariateMixture.from_parameters(*args, **kwargs)

    def get_xmargin(self, log=True):
        pdf = np.zeros_like(self.supportx)
        for i in range(self.n_components):
            pdf += self.get_component_marginal_pdf(i, 0, True)
        if not log:
            support = np.exp(self.supportx)
            pdf /= support
        else:
            support = self.supportx
        return support, pdf

    def get_ymargin(self, log=True):
        pdf = np.zeros_like(self.supporty)
        for i in range(self.n_components):
            pdf += self.get_component_marginal_pdf(i, 1, True)
        if not log:
            support = np.exp(self.supporty)
            pdf /= support
        else:
            support = self.supporty
        return support, pdf

    def get_component_marginal_pdf(self, idx, margin, weighted=True):
        component = self.components[idx]
        mu = component.mean[margin]
        sigma = np.sqrt(component.cov[margin, margin])
        pdf = norm(mu, sigma).pdf(self.supportx)

        if weighted:
            pdf *= self.weights_[idx]

        return pdf




from unittest.mock import patch

from os.path import join, exists
from pandas import DataFrame

from flyqma.data import Stack
from flyqma.data.layers import Layer
from .test_io import TestPaths


class Test02_Stack(TestPaths):
    """
    Tests for Stack instance.
    """

    @classmethod
    def setUpClass(cls):
        """ Initialize test instance of Stack. """
        with patch('builtins.input', side_effect=['yes', 12]):
            stack = Stack(cls.stack_path, bit_depth=cls.bit_depth)
        cls.stack = stack

    @classmethod
    def tearDownClass(cls):
        """ Restore stack directory to original state. """
        cls.stack.restore_directory()

    def test01_paths(self):
        """ Check that child directories are present. """
        tests = [exists(self.stack.layers_path),
                 exists(self.stack.path)]
        self.assertTrue(False not in tests)

    def test02_load_image(self):
        """ Load stack image. """
        self.stack.load_image()
        self.assertTrue(self.stack.stack is not None)

    def test03_load_layer(self):
        """ Load Layer from stack. """
        layer = self.stack.load_layer()
        self.assertTrue(isinstance(layer, Layer))

    def test04_segmentation(self):
        """ Segment all layers in stack.. """
        try:
            self.stack.segment(2)
            success = True
        except:
            success = False
        self.assertTrue(success)

    def test05_train_annotator(self):
        """ Train clone annotation model on entire stack. """
        try:
            self.stack.train_annotator('ch0', save=True, max_num_components=3)
            success = True
        except:
            success = False
        self.assertTrue(success)

    def test06_apply_annotator(self):
        """ Annotate a single layer within the stack. """
        layer = self.stack[0]
        layer.build_graph('ch0')
        layer.annotate()
        self.assertTrue('genotype' in layer.data.columns)

    def test07_aggregate_raw_data(self):
        """ Collect raw measurements from all layers. """
        data = self.stack.aggregate_measurements(raw=True)
        self.assertTrue(isinstance(data, DataFrame))

    def test07_aggregate_data_excluding_boundary(self):
        """ Collect measurements from all layers, excluding boundaries. """
        data = self.stack.aggregate_measurements(exclude_boundary=True)
        self.assertTrue(isinstance(data, DataFrame))



import numpy as np
import statsmodels.api as sm
import warnings


class OLS:
    """
    Ordinary least squares model fit to X and Y data.

    Attributes:

        x, y (array like) - data

        model (sm.OLS) - fitted OLS model

    """

    def __init__(self, x, y, **fit_kw):
        """ Instantiate ordinary least squared model. """
        self.x = x
        self.y = y
        self.fit(**fit_kw)

    def fit(self, **kw):
        """ Fit model to X and Y data. """
        self.domain = np.linspace(0, self.x.max(), 10)
        x = sm.tools.add_constant(self.x.reshape(-1, 1))
        self.model = sm.OLS(self.y, x, hasconst=None).fit()

    def predict(self, x):
        """ Make model prediction. """
        xx = sm.tools.add_constant(x.reshape(-1, 1))
        return self.model.predict(xx)

    def detrend(self, x, y):
        """ Remove linear trend from X and Y data. """
        return y - self.predict(x)


class GLM(OLS):
    """
    Generalized linear model with gamma distributed residuals and an identity link function fit to X and Y data.

    Attributes:

        model (sm.genmod.GLM) - generalized linear model

        domain (np.ndarray[float]) - regularly spaced x-domain

    Inherited attributes:

        x, y (array like) - data

    """

    def fit(self, N=100000, maxiter=500, shift=0):
        """
        Fit Gamma GLM with identity link function.

        Args:

            N (int) - number of samples used

            maxiter (int) - maximum number of iterations for optimization

            shift (float) - offset used to keep values positive

        """

        self.domain = np.linspace(0, self.x.max(), 10)

        # downsample
        if N is not None:
            ind = np.random.randint(0, self.x.size, size=N)
        else:
            ind = np.arange(self.x.size)

        x, y = self.x[ind], self.y[ind]

        # construct variables
        xx = sm.tools.add_constant(x.reshape(-1, 1))
        yy = y + shift

        # define model
        family = sm.families.Gamma(link=sm.families.links.identity())
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            glm = sm.genmod.GLM(yy, xx, family=family)

        # fit model
        start_params = [0.1+shift, 0.5]
        self.model = glm.fit(start_params=start_params, maxiter=maxiter)
        if not self.model.converged:
            raise Warning('GLM did not converge.')
